# Unix SMB/CIFS implementation.
# Copyright (C) Andrew Bartlett <abartlet@samba.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from samba import provision, param
import os
import random
import shutil
import subprocess
from samba.tests import (env_loadparm, create_test_ou, BlackboxProcessError,
                         BlackboxTestCase, connect_samdb, delete_force)
from samba.tests.gkdi import GkdiBaseTest
import ldb
from samba.samdb import SamDB
from samba.auth import system_session
from samba import Ldb, dn_from_dns_name
from samba.netcmd import CommandError
from samba.netcmd.fsmo import get_fsmo_roleowner
import re
from samba import sites
from samba.dsdb import (
    _dsdb_load_udv_v2,
    DS_GUID_DOMAIN_CONTROLLERS_CONTAINER,
)
from samba import dsdb
from samba import safe_tarfile as tarfile


def get_prim_dom(secrets_path, lp):
    secrets_ldb = Ldb(secrets_path, session_info=system_session(), lp=lp)
    return secrets_ldb.search(base="CN=Primary Domains",
                              attrs=['objectClass', 'samAccountName',
                                     'secret', 'msDS-KeyVersionNumber'],
                              scope=ldb.SCOPE_SUBTREE,
                              expression="(objectClass=kerberosSecret)")

# A set of attributes we know should be confidential.
confidential_attrs = frozenset({
    "msFVE-KeyPackage",
    "msFVE-RecoveryPassword",
    "msKds-RootKeyData",
})


def confidential_values_present(msg: ldb.Message) -> bool:
    return any(map(msg.get, confidential_attrs))


# The backup tests require that a completely clean LoadParm object gets used
# for the restore. Otherwise the same global LP gets re-used, and the LP
# settings can bleed from one test case to another.
# To do this, these tests should use check_output(), which executes the command
# in a separate process (as opposed to runcmd(), runsubcmd()).
# So although this is a samba-tool test, we don't inherit from SambaToolCmdTest
# so that we never inadvertently use .runcmd() by accident.
class DomainBackupBase(BlackboxTestCase, GkdiBaseTest):

    def setUp(self):
        super().setUp()

        server = os.environ["DC_SERVER"]
        self.user_auth = "-U%s%%%s" % (os.environ["DC_USERNAME"],
                                       os.environ["DC_PASSWORD"])

        # LDB connection to the original server being backed up
        self.ldb = connect_samdb("ldap://%s" % server)
        self.new_server = "BACKUPSERV"
        self.server = server.upper()
        self.base_cmd = None
        self.backup_markers = ['sidForRestore', 'backupDate']
        self.restore_domain = os.environ["DOMAIN"]
        self.restore_realm = os.environ["REALM"]
        self.backend = None

    def use_backend(self, backend):
        """Explicitly set the DB backend that the backup should use"""
        self.backend = backend
        self.base_cmd += ["--backend-store=" + backend]

    def get_expected_partitions(self, samdb):
        basedn = str(samdb.get_default_basedn())
        config_dn = "CN=Configuration,%s" % basedn
        return [basedn, config_dn, "CN=Schema,%s" % config_dn,
                "DC=DomainDnsZones,%s" % basedn,
                "DC=ForestDnsZones,%s" % basedn]

    def assert_partitions_present(self, samdb):
        """Asserts all expected partitions are present in the backup samdb"""
        res = samdb.search(base="", scope=ldb.SCOPE_BASE,
                           attrs=['namingContexts'])
        actual_ncs = [str(r) for r in res[0].get('namingContexts')]

        expected_ncs = self.get_expected_partitions(samdb)

        for nc in expected_ncs:
            self.assertTrue(nc in actual_ncs,
                            "%s not in %s" % (nc, str(actual_ncs)))

    def assert_repl_uptodate_vector(self, samdb):
        """Asserts an replUpToDateVector entry exists for the original DC"""
        orig_invoc_id = self.ldb.get_invocation_id()
        expected_ncs = self.get_expected_partitions(samdb)

        # loop through the partitions and check the upToDateness vector
        for nc in expected_ncs:
            found = False
            for cursor in _dsdb_load_udv_v2(samdb, nc):
                if orig_invoc_id == str(cursor.source_dsa_invocation_id):
                    found = True
                    break
            self.assertTrue(found, "Couldn't find UDTV for original DC")

    def assert_dcs_present(self, samdb, expected_server, expected_count=None):
        """Checks that the expected server is present in the restored DB"""
        search_expr = "(&(objectClass=Server)(serverReference=*))"
        res = samdb.search(samdb.get_config_basedn(),
                           scope=ldb.SCOPE_SUBTREE,
                           expression=search_expr)
        server_found = False
        for msg in res:
            if expected_server in str(msg.dn):
                server_found = True

        self.assertTrue(server_found,
                        "Could not find %s server" % expected_server)

        if expected_count is not None:
            self.assertTrue(len(res) == expected_count)

    def restore_dir(self):
        extract_dir = os.path.join(self.tempdir, 'tree')
        if not os.path.exists(extract_dir):
            os.mkdir(extract_dir)
            self.addCleanup(shutil.rmtree, extract_dir)
        return extract_dir

    def untar_backup(self, backup_file):
        """Untar the backup file's raw contents (i.e. not a proper restore)"""
        extract_dir = self.restore_dir()
        with tarfile.open(backup_file) as tf:
            tf.extractall(extract_dir)
        return extract_dir

    def _test_backup_untar(
        self,
        primary_domain_secrets=0,
        verify_checksums=False
    ):
        """Creates a backup, untars the raw files, and sanity-checks the DB"""
        backup_file = self.create_backup()
        extract_dir = self.untar_backup(backup_file)

        if (verify_checksums):
            p = subprocess.Popen(
                ["sha256sum", "-c", "SHA256SUM"],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                cwd=extract_dir,
            )
            (out, err) = p.communicate()
            if p.returncode:
                print("Error: " + err.decode('utf-8'))
                raise CommandError('Failed to verify checksums')

        private_dir = os.path.join(extract_dir, "private")
        samdb_path = os.path.join(private_dir, "sam.ldb")
        lp = env_loadparm()
        samdb = SamDB(url=samdb_path, session_info=system_session(), lp=lp)

        # check that backup markers were added to the DB
        res = samdb.search(base=ldb.Dn(samdb, "@SAMBA_DSDB"),
                           scope=ldb.SCOPE_BASE,
                           attrs=self.backup_markers)
        self.assertEqual(len(res), 1)
        for marker in self.backup_markers:
            self.assertIsNotNone(res[0].get(marker),
                                 "%s backup marker missing" % marker)

        # check the secrets.ldb entry for the primary domain. (Online/clone
        # backups shouldn't have this, as they never got it during the backup)
        secrets_path = os.path.join(private_dir, "secrets.ldb")
        res = get_prim_dom(secrets_path, lp)
        self.assertEqual(len(res), primary_domain_secrets)

        # sanity-check that all the partitions got backed up
        self.assert_partitions_present(samdb)

    def _test_backup_restore(self):
        """Does a backup/restore, with specific checks of the resulting DB"""
        backup_file = self.create_backup()
        self.restore_backup(backup_file)
        lp = self.check_restored_smbconf()
        restored_samdb = self.check_restored_database(lp)

        # Search across naming contexts to ensure we don’t miss any objects.
        cross_ncs = ["search_options:1:2"]

        # Search the entire restored database for GKDI root keys.
        restored_root_keys = restored_samdb.search(
            restored_samdb.get_root_basedn(),
            expression="objectClass=msKds-ProvRootKey",
            scope=ldb.SCOPE_SUBTREE,
            attrs=list(confidential_attrs),
            controls=cross_ncs,
        )
        # Assert that at least one key exists.
        self.assertLess(0, len(restored_root_keys))

        # Assert that the confidential material is still present.
        for msg in restored_root_keys:
            self.assertTrue(confidential_values_present(msg))

    def _test_backup_restore_no_secrets(self):
        """Does a backup/restore with secrets excluded from the resulting DB"""

        samdb: SamDB = self.ldb

        # Use the DC account as the parent of a new msFVE-RecoveryInformation object.
        mach_dn = samdb.get_wellknown_dn(
            samdb.get_default_basedn(), DS_GUID_DOMAIN_CONTROLLERS_CONTAINER
        )
        mach_dn.add_child(f"CN={self.server}")

        # Get the machine account's GUID.
        res = samdb.search(mach_dn, scope=ldb.SCOPE_BASE, attrs=["objectGUID"])
        mach_guid = res[0].get("objectGUID", idx=0)

        # Create a child msFVE-RecoveryInformation object with some confidential
        # attributes.

        recovery_dn = ldb.Dn(samdb, str(mach_dn))
        recovery_dn.add_child(f"CN=recovery_info_{random.randint(0, 0xFFFF)}")

        self.addCleanup(delete_force, samdb, recovery_dn)
        samdb.add({
            "dn": recovery_dn,
            "objectClass": "msFVE-RecoveryInformation",
            "msFVE-RecoveryGuid": mach_guid,
            "msFVE-KeyPackage": b"secret key package",
            "msFVE-RecoveryPassword": "Secret007",
        })

        def ldap_display_name(item: ldb.Message) -> str:
            return item.get("lDAPDisplayName", idx=0).decode()

        # Look up all the confidential attributes in the schema.
        res = samdb.search(
            samdb.get_schema_basedn(),
            expression=(
                f"(&(objectClass=attributeSchema)"
                f"(searchFlags:{ldb.OID_COMPARATOR_AND}"
                f":={dsdb.SEARCH_FLAG_CONFIDENTIAL}))"
            ),
            attrs=["lDAPDisplayName"],
        )
        all_confidential_attrs = set(map(ldap_display_name, res))

        # Ensure that the attributes we think are confidential actually are.
        self.assertLessEqual(confidential_attrs, all_confidential_attrs)

        # Search for GKDI root keys containing secret material.
        root_key_container_dn = self.get_root_key_container_dn(samdb)
        root_keys = samdb.search(
            root_key_container_dn,
            scope=ldb.SCOPE_ONELEVEL,
            attrs=list(confidential_attrs),
        )
        # Assert that at least one key exists.
        self.assertLess(0, len(root_keys))

        # Assert that the root keys all have confidential material.
        for msg in root_keys:
            self.assertTrue(confidential_values_present(msg))

        # Assert that the recovery information object has confidential material.
        recovery_info = samdb.search(
            recovery_dn, scope=ldb.SCOPE_BASE, attrs=list(confidential_attrs)
        )
        self.assertTrue(confidential_values_present(recovery_info[0]))

        # We don’t need the original samdb any more.
        del samdb

        # exclude secrets when we create the backup
        backup_file = self.create_backup(extra_args=["--no-secrets"])
        self.restore_backup(backup_file)
        lp = self.check_restored_smbconf()

        # assert that we don't find user secrets in the DB
        restored_samdb = self.check_restored_database(lp, expect_secrets=False)

        # For the following searches, search across naming contexts to ensure we
        # don’t miss any objects.
        cross_ncs = ["search_options:1:2"]

        # Search the entire restored database for GKDI root keys.
        restored_root_keys = restored_samdb.search(
            restored_samdb.get_root_basedn(),
            expression="objectClass=msKds-ProvRootKey",
            scope=ldb.SCOPE_SUBTREE,
            attrs=list(confidential_attrs),
            controls=cross_ncs,
        )
        # Assert that no keys exist.
        self.assertEqual(0, len(restored_root_keys))

        # Search the entire restored database for recovery info objects.
        recovery_info_objs = restored_samdb.search(
            restored_samdb.get_root_basedn(),
            expression="objectClass=msFVE-RecoveryInformation",
            scope=ldb.SCOPE_SUBTREE,
            attrs=list(confidential_attrs),
            controls=cross_ncs,
        )
        # Assert that no recovery info exists.
        self.assertEqual(0, len(recovery_info_objs))

    def _test_backup_restore_into_site(self):
        """Does a backup and restores into a non-default site"""

        # create a new non-default site
        sitename = "Test-Site-For-Backups"
        sites.create_site(self.ldb, self.ldb.get_config_basedn(), sitename)
        self.addCleanup(sites.delete_site, self.ldb,
                        self.ldb.get_config_basedn(), sitename)

        # restore the backup DC into the site we just created
        backup_file = self.create_backup()
        self.restore_backup(backup_file, ["--site=" + sitename])

        lp = self.check_restored_smbconf()
        restored_ldb = self.check_restored_database(lp)

        # check the restored DC was added to the site we created, i.e. there's
        # an entry matching the new DC sitting underneath the site DN
        site_dn = "CN={0},CN=Sites,{1}".format(sitename,
                                               restored_ldb.get_config_basedn())
        match_server = "(&(objectClass=server)(cn={0}))".format(self.new_server)
        res = restored_ldb.search(site_dn, scope=ldb.SCOPE_SUBTREE,
                                  expression=match_server)
        self.assertTrue(len(res) == 1,
                        "Failed to find new DC under site")

    def create_smbconf(self, settings):
        """Creates a very basic smb.conf to pass to the restore tool"""

        # without the testenv config's settings, the NTACL backup_restore()
        # operation will fail (because we're not root). So first suck in all
        # testenv's settings, so we retain these in the new config. Note we
        # use a non-global LP so that these settings don't leak into other
        # places we use LoadParms
        testenv_conf = os.environ["SMB_CONF_PATH"]
        local_lp = param.LoadParm(filename_for_non_global_lp=testenv_conf)

        # add the new settings to the LP, then write the settings to file
        for key, val in settings.items():
            local_lp.set(key, val)

        new_smbconf = os.path.join(self.tempdir, "smb.conf")
        local_lp.dump(False, new_smbconf)

        self.addCleanup(os.remove, new_smbconf)
        return new_smbconf

    def _test_backup_restore_with_conf(self):
        """Checks smb.conf values passed to the restore are retained"""
        backup_file = self.create_backup()

        # create an smb.conf that we pass to the restore. The netbios/state
        # dir should get overridden by the restore, the other settings should
        # trickle through into the restored dir's smb.conf
        settings = {'state directory': '/var/run',
                    'netbios name': 'FOOBAR',
                    'workgroup': 'NOTMYDOMAIN',
                    'realm': 'NOT.MY.REALM'}
        assert_settings = {'drs: max link sync': '275',
                           'prefork children': '7'}
        settings.update(assert_settings)
        smbconf = self.create_smbconf(settings)

        self.restore_backup(backup_file, ["--configfile=" + smbconf])

        # this will check netbios name/state dir
        lp = self.check_restored_smbconf()
        self.check_restored_database(lp)

        # check the remaining settings are still intact
        for key, val in assert_settings.items():
            self.assertEqual(str(lp.get(key)), val,
                             "'%s' was '%s' in smb.conf" % (key, lp.get(key)))

    def check_restored_smbconf(self):
        """Sanity-check important smb.conf values are restored correctly"""
        smbconf = os.path.join(self.restore_dir(), "etc", "smb.conf")
        bkp_lp = param.LoadParm(filename_for_non_global_lp=smbconf)
        self.assertEqual(bkp_lp.get('netbios name'), self.new_server)
        self.assertEqual(bkp_lp.get('workgroup'), self.restore_domain)
        self.assertEqual(bkp_lp.get('realm'), self.restore_realm.upper())

        # we restore with a fixed directory structure, so we can sanity-check
        # that the core filepaths settings are what we expect them to be
        private_dir = os.path.join(self.restore_dir(), "private")
        self.assertEqual(bkp_lp.get('private dir'), private_dir)
        state_dir = os.path.join(self.restore_dir(), "state")
        self.assertEqual(bkp_lp.get('state directory'), state_dir)
        return bkp_lp

    def check_restored_database(self, bkp_lp, expect_secrets=True):
        paths = provision.provision_paths_from_lp(bkp_lp, bkp_lp.get("realm"))

        bkp_pd = get_prim_dom(paths.secrets, bkp_lp)
        self.assertEqual(len(bkp_pd), 1)
        account = bkp_pd[0].get('samAccountName')
        self.assertIsNotNone(account)
        self.assertEqual(str(account[0]), self.new_server + '$')
        self.assertIsNotNone(bkp_pd[0].get('secret'))

        samdb = SamDB(url=paths.samdb, session_info=system_session(),
                      lp=bkp_lp, credentials=self.get_credentials())

        # check that the backup markers have been removed from the restored DB
        res = samdb.search(base=ldb.Dn(samdb, "@SAMBA_DSDB"),
                           scope=ldb.SCOPE_BASE,
                           attrs=self.backup_markers)
        self.assertEqual(len(res), 1)
        for marker in self.backup_markers:
            self.assertIsNone(res[0].get(marker),
                              "%s backup-marker left behind" % marker)

        # check that the repsFrom and repsTo values have been removed
        # from the restored DB
        res = samdb.search(base=samdb.get_default_basedn(),
                           scope=ldb.SCOPE_BASE,
                           attrs=['repsFrom', 'repsTo'])
        self.assertEqual(len(res), 1)
        self.assertIsNone(res[0].get('repsFrom'))
        self.assertIsNone(res[0].get('repsTo'))

        res = samdb.search(base=samdb.get_config_basedn(),
                           scope=ldb.SCOPE_BASE,
                           attrs=['repsFrom', 'repsTo'])
        self.assertEqual(len(res), 1)
        self.assertIsNone(res[0].get('repsFrom'))
        self.assertIsNone(res[0].get('repsTo'))

        # check the DB is using the backend we supplied
        if self.backend:
            res = samdb.search(base="@PARTITION", scope=ldb.SCOPE_BASE,
                               attrs=["backendStore"])
            backend = str(res[0].get("backendStore"))
            self.assertEqual(backend, self.backend)

        # check the restored DB has the expected partitions/DC/FSMO roles
        self.assert_partitions_present(samdb)
        self.assert_dcs_present(samdb, self.new_server, expected_count=1)
        self.assert_fsmo_roles(samdb, self.new_server, self.server)
        self.assert_secrets(samdb, expect_secrets=expect_secrets)

        # check we still have an uptodateness vector for the original DC
        self.assert_repl_uptodate_vector(samdb)
        return samdb

    def assert_user_secrets(self, samdb, username, expect_secrets):
        """Asserts that a user has/doesn't have secrets as expected"""
        basedn = str(samdb.get_default_basedn())
        user_dn = "CN=%s,CN=users,%s" % (username, basedn)

        if expect_secrets:
            self.assertIsNotNone(samdb.searchone("unicodePwd", user_dn))
        else:
            # the search should throw an exception because the secrets
            # attribute isn't actually there
            self.assertRaises(KeyError, samdb.searchone, "unicodePwd", user_dn)

    def assert_secrets(self, samdb, expect_secrets):
        """Check the secrets in the restored DB match what's expected"""

        # check secrets for the built-in testenv users match what's expected
        test_users = ["alice", "bob", "jane", "joe"]
        for user in test_users:
            self.assert_user_secrets(samdb, user, expect_secrets)

    def assert_fsmo_roles(self, samdb, server, exclude_server):
        """Asserts the expected server is the FSMO role owner"""
        domain_dn = samdb.domain_dn()
        forest_dn = dn_from_dns_name(samdb.forest_dns_name())
        fsmos = {'infrastructure': "CN=Infrastructure," + domain_dn,
                 'naming': "CN=Partitions,%s" % samdb.get_config_basedn(),
                 'schema': str(samdb.get_schema_basedn()),
                 'rid': "CN=RID Manager$,CN=System," + domain_dn,
                 'pdc': domain_dn,
                 'domaindns':
                     "CN=Infrastructure,DC=DomainDnsZones," + domain_dn,
                 'forestdns':
                     "CN=Infrastructure,DC=ForestDnsZones," + forest_dn}
        for role, dn in fsmos.items():
            owner = get_fsmo_roleowner(samdb, ldb.Dn(samdb, dn), role)
            self.assertTrue("CN={0},".format(server) in owner.extended_str(),
                            "Expected %s to own FSMO role %s" % (server, role))
            self.assertTrue("CN={0},".format(exclude_server)
                            not in owner.extended_str(),
                            "%s found as FSMO %s role owner" % (server, role))

    def cleanup_tempdir(self):
        for filename in os.listdir(self.tempdir):
            filepath = os.path.join(self.tempdir, filename)
            if os.path.isfile(filepath):
                os.remove(filepath)
            elif os.path.isdir(filepath):
                shutil.rmtree(filepath)

    def run_cmd(self, args):
        """Executes a samba-tool backup/restore command"""

        cmd = " ".join(args)
        print("Executing: samba-tool %s" % cmd)
        try:
            # note: it's important we run the cmd in a separate process here
            out = self.check_output("samba-tool " + cmd)
        except BlackboxProcessError as e:
            # if the command failed, it may have left behind temporary files.
            # We're going to fail the test, but first cleanup any temp files so
            # that we skip the TestCaseInTempDir._remove_tempdir() assertions
            self.cleanup_tempdir()
            self.fail("Error calling samba-tool: %s" % e)
        print(out.decode())

    def create_backup(self, extra_args=None):
        """Runs the backup cmd to produce a backup file for the testenv DC"""
        # Run the backup command and check we got one backup tar file
        args = self.base_cmd + ["--targetdir=" + self.tempdir]
        if extra_args:
            args += extra_args

        self.run_cmd(args)

        # find the filename of the backup-file generated
        tar_files = []
        for fn in os.listdir(self.tempdir):
            if (fn.startswith("samba-backup-") and fn.endswith(".tar.bz2")):
                tar_files.append(fn)

        self.assertTrue(len(tar_files) == 1,
                        "Domain backup created %u tar files" % len(tar_files))

        # clean up the backup file once the test finishes
        backup_file = os.path.join(self.tempdir, tar_files[0])
        self.addCleanup(os.remove, backup_file)
        return backup_file

    def restore_backup(self, backup_file, extra_args=None):
        """Restores the samba directory files from a given backup"""
        # Run the restore command
        extract_dir = self.restore_dir()
        args = ["domain", "backup", "restore", "--backup-file=" + backup_file,
                "--targetdir=" + extract_dir,
                "--newservername=" + self.new_server]
        if extra_args:
            args += extra_args

        self.run_cmd(args)

        # sanity-check the restore doesn't modify the original DC by mistake
        self.assert_partitions_present(self.ldb)
        self.assert_dcs_present(self.ldb, self.server)
        self.assert_fsmo_roles(self.ldb, self.server, self.new_server)


class DomainBackupOnline(DomainBackupBase):

    def setUp(self):
        super().setUp()
        self.base_cmd = ["domain", "backup", "online",
                         "--server=" + self.server, self.user_auth]

    # run the common test cases above using online backups
    def test_backup_untar(self):
        self._test_backup_untar()

    def test_backup_restore(self):
        self.use_backend("tdb")
        self._test_backup_restore()

    def test_backup_restore_with_conf(self):
        self.use_backend("mdb")
        self._test_backup_restore_with_conf()

    def test_backup_restore_no_secrets(self):
        self.use_backend("tdb")
        self._test_backup_restore_no_secrets()

    def test_backup_restore_into_site(self):
        self.use_backend("mdb")
        self._test_backup_restore_into_site()


class DomainBackupRename(DomainBackupBase):

    # run the above test cases using a rename backup
    def setUp(self):
        super().setUp()
        self.new_server = "RENAMESERV"
        self.restore_domain = "NEWDOMAIN"
        self.restore_realm = "rename.test.net"
        self.new_basedn = "DC=rename,DC=test,DC=net"
        self.base_cmd = ["domain", "backup", "rename", self.restore_domain,
                         self.restore_realm, "--server=" + self.server,
                         self.user_auth]
        self.backup_markers += ['backupRename']

    # run the common test case code for backup-renames
    def test_backup_untar(self):
        self._test_backup_untar()

    def test_backup_restore(self):
        self.use_backend("mdb")
        self._test_backup_restore()

    def test_backup_restore_with_conf(self):
        self.use_backend("tdb")
        self._test_backup_restore_with_conf()

    def test_backup_restore_no_secrets(self):
        self.use_backend("mdb")
        self._test_backup_restore_no_secrets()

    def test_backup_restore_into_site(self):
        self.use_backend("tdb")
        self._test_backup_restore_into_site()

    def test_backup_invalid_args(self):
        """Checks that rename commands with invalid args are rejected"""

        # try a "rename" using the same realm as the DC currently has
        rename_cmd = "samba-tool domain backup rename "
        bad_cmd = "{cmd} {domain} {realm}".format(cmd=rename_cmd,
                                                  domain=self.restore_domain,
                                                  realm=os.environ["REALM"])
        self.assertRaises(BlackboxProcessError, self.check_output, bad_cmd)

        # try a "rename" using the same domain as the DC currently has
        bad_cmd = "{cmd} {domain} {realm}".format(cmd=rename_cmd,
                                                  domain=os.environ["DOMAIN"],
                                                  realm=self.restore_realm)
        self.assertRaises(BlackboxProcessError, self.check_output, bad_cmd)

    def add_link(self, attr, source, target):
        m = ldb.Message()
        m.dn = ldb.Dn(self.ldb, source)
        m[attr] = ldb.MessageElement(target, ldb.FLAG_MOD_ADD, attr)
        self.ldb.modify(m)

    def test_one_way_links(self):
        """Sanity-check that a rename handles one-way links correctly"""

        # Do some initial setup on the DC before back it up:
        # create an OU to hold the test objects we'll create
        test_ou = create_test_ou(self.ldb, "rename_test")
        self.addCleanup(self.ldb.delete, test_ou, ["tree_delete:1"])

        # create the source and target objects and link them together.
        # We use addressBookRoots2 here because it's a one-way link
        src_dn = "CN=link_src,%s" % test_ou
        self.ldb.add({"dn": src_dn,
                      "objectclass": "msExchConfigurationContainer"})
        target_dn = "OU=link_tgt,%s" % test_ou
        self.ldb.add({"dn": target_dn, "objectclass": "organizationalunit"})
        link_attr = "addressBookRoots2"
        self.add_link(link_attr, src_dn, target_dn)

        # add a second link target that's in a different partition
        server_dn = ("CN=testrename,CN=Servers,CN=Default-First-Site-Name,"
                     "CN=Sites,%s" % str(self.ldb.get_config_basedn()))
        self.ldb.add({"dn": server_dn, "objectclass": "server"})
        self.addCleanup(self.ldb.delete, server_dn)
        self.add_link(link_attr, src_dn, server_dn)

        # do the backup/restore
        backup_file = self.create_backup()
        self.restore_backup(backup_file)
        lp = self.check_restored_smbconf()
        restored_ldb = self.check_restored_database(lp)

        # work out what the new DNs should be
        old_basedn = str(self.ldb.get_default_basedn())
        new_target_dn = re.sub(old_basedn + '$', self.new_basedn, target_dn)
        new_src_dn = re.sub(old_basedn + '$', self.new_basedn, src_dn)
        new_server_dn = re.sub(old_basedn + '$', self.new_basedn, server_dn)

        # check the links exist in the renamed DB with the correct DNs
        res = restored_ldb.search(base=new_src_dn, scope=ldb.SCOPE_BASE,
                                  attrs=[link_attr])
        self.assertEqual(len(res), 1,
                         "Failed to find renamed link source object")
        self.assertTrue(link_attr in res[0], "Missing link attribute")
        link_values = [str(x) for x in res[0][link_attr]]
        self.assertTrue(new_target_dn in link_values)
        self.assertTrue(new_server_dn in link_values)

    # extra checks we run on the restored DB in the rename case
    def check_restored_database(self, bkp_lp, expect_secrets=True):
        # run the common checks over the restored DB
        common_test = super()
        samdb = common_test.check_restored_database(bkp_lp, expect_secrets)

        # check we have actually renamed the DNs
        basedn = str(samdb.get_default_basedn())
        self.assertEqual(basedn, self.new_basedn)

        # check the partition and netBIOS name match the new domain
        partitions_dn = samdb.get_partitions_dn()
        nc_name = ldb.binary_encode(str(basedn))
        res = samdb.search(base=partitions_dn, scope=ldb.SCOPE_ONELEVEL,
                           attrs=["nETBIOSName", "cn"],
                           expression='ncName=%s' % nc_name)
        self.assertEqual(len(res), 1,
                         "Looking up partition's NetBIOS name failed")
        self.assertEqual(str(res[0].get("nETBIOSName")), self.restore_domain)
        self.assertEqual(str(res[0].get("cn")), self.restore_domain)

        # check the DC has the correct dnsHostname
        realm = self.restore_realm
        dn = "CN=%s,OU=Domain Controllers,%s" % (self.new_server,
                                                 self.new_basedn)
        res = samdb.search(base=dn, scope=ldb.SCOPE_BASE,
                           attrs=["dNSHostName"])
        self.assertEqual(len(res), 1,
                         "Looking up new DC's dnsHostname failed")
        expected_val = "%s.%s" % (self.new_server.lower(), realm)
        self.assertEqual(str(res[0].get("dNSHostName")), expected_val)

        # check the DNS zones for the new realm are present
        dn = "DC=%s,CN=MicrosoftDNS,DC=DomainDnsZones,%s" % (realm, basedn)
        res = samdb.search(base=dn, scope=ldb.SCOPE_BASE)
        self.assertEqual(len(res), 1, "Lookup of new domain's DNS zone failed")

        forestdn = samdb.get_root_basedn().get_linearized()
        dn = "DC=_msdcs.%s,CN=MicrosoftDNS,DC=ForestDnsZones,%s" % (realm,
                                                                    forestdn)
        res = samdb.search(base=dn, scope=ldb.SCOPE_BASE)
        self.assertEqual(len(res), 1, "Lookup of new domain's DNS zone failed")
        return samdb


class DomainBackupOffline(DomainBackupBase):

    def setUp(self):
        super().setUp()
        self.base_cmd = ["domain", "backup", "offline"]

    def test_backup_untar(self):
        self._test_backup_untar(primary_domain_secrets=1, verify_checksums=True)

    def test_backup_restore_with_conf(self):
        self._test_backup_restore_with_conf()

    def test_backup_restore(self):
        self._test_backup_restore()

    def test_backup_restore_into_site(self):
        self._test_backup_restore_into_site()



# Unix SMB/CIFS implementation.
#
# Copyright (C) Samuel Cabrero <scabrero@suse.de> 2018
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
from subprocess import Popen, PIPE
from samba.tests import BlackboxProcessError
from samba.tests.ntlm_auth_base import NTLMAuthTestCase
from samba.common import get_string

class NTLMAuthHelpersTests(NTLMAuthTestCase):

    def setUp(self):
        super().setUp()
        self.username = os.environ["DC_USERNAME"]
        self.password = os.environ["DC_PASSWORD"]
        self.domain = os.environ["DOMAIN"]
        out = get_string(self.check_output("wbinfo -n %s" % self.username))
        self.group_sid = out.split(" ")[0]
        self.assertTrue(self.group_sid.startswith("S-1-5-21-"))
        self.bad_group_sid = self.group_sid[:-2]

    def test_specified_domain(self):
        """ ntlm_auth with specified domain """

        username = "foo"
        password = "secret"
        domain = "FOO"

        ret = self.run_helper(client_username=username,
                              client_password=password,
                              client_domain=domain,
                              server_username=username,
                              server_password=password,
                              server_domain=domain,
                              server_use_winbind=False)
        self.assertTrue(ret)

        username = "foo"
        password = "secret"
        domain = "fOo"

        ret = self.run_helper(client_username=username,
                              client_password=password,
                              client_domain=domain,
                              server_username=username,
                              server_password=password,
                              server_domain=domain,
                              server_use_winbind=False)
        self.assertTrue(ret)

    def test_against_winbind(self):
        """ ntlm_auth against winbindd """

        ret = self.run_helper(client_username=self.username,
                              client_password=self.password,
                              client_domain=self.domain,
                              server_use_winbind=True)
        self.assertTrue(ret)

    def test_ntlmssp_gss_spnego(self):
        """ ntlm_auth with NTLMSSP client and gss-spnego server """

        username = "foo"
        password = "secret"
        domain = "fOo"

        ret = self.run_helper(client_username=username,
                              client_password=password,
                              client_domain=domain,
                              server_username=username,
                              server_password=password,
                              server_domain=domain,
                              client_helper="ntlmssp-client-1",
                              server_helper="gss-spnego",
                              server_use_winbind=False)
        self.assertTrue(ret)

    def test_gss_spnego(self):
        """ ntlm_auth with NTLMSSP gss-spnego-client and gss-spnego server """

        username = "foo"
        password = "secret"
        domain = "fOo"

        ret = self.run_helper(client_username=username,
                              client_password=password,
                              client_domain=domain,
                              server_username=username,
                              server_password=password,
                              server_domain=domain,
                              client_helper="gss-spnego-client",
                              server_helper="gss-spnego",
                              server_use_winbind=False)
        self.assertTrue(ret)

    def test_gss_spnego_winbind(self):
        """ ntlm_auth with NTLMSSP gss-spnego-client and gss-spnego server
        against winbind """

        ret = self.run_helper(client_username=self.username,
                              client_password=self.password,
                              client_domain=self.domain,
                              client_helper="gss-spnego-client",
                              server_helper="gss-spnego",
                              server_use_winbind=True)
        self.assertTrue(ret)

    def test_ntlmssp_gss_spnego_cached_creds(self):
        """ ntlm_auth with NTLMSSP client and gss-spnego server against
        winbind with cached credentials """

        param = "--ccache-save=%s%s%s%%%s" % (self.domain,
                                              self.winbind_separator,
                                              self.username,
                                              self.password)
        cache_cmd = ["wbinfo",
                     param]
        self.check_exit_code(cache_cmd, 0)

        ret = self.run_helper(client_username=self.username,
                              client_password=None,
                              client_domain=self.domain,
                              client_use_cached_creds=True,
                              client_helper="ntlmssp-client-1",
                              server_helper="gss-spnego",
                              server_use_winbind=True)
        self.assertTrue(ret)

    def test_require_membership(self):
        """ ntlm_auth against winbindd with require-membership-of """

        ret = self.run_helper(client_username=self.username,
                              client_password=self.password,
                              client_domain=self.domain,
                              require_membership=self.group_sid,
                              server_use_winbind=True)
        self.assertTrue(ret)

        ret = self.run_helper(client_username=self.username,
                              client_password=self.password,
                              client_domain=self.domain,
                              require_membership=self.bad_group_sid,
                              server_use_winbind=True)
        self.assertFalse(ret)

    def test_require_membership_gss_spnego(self):
        """ ntlm_auth with NTLMSSP gss-spnego-client and gss-spnego server
        against winbind with require-membership-of """

        ret = self.run_helper(client_username=self.username,
                              client_password=self.password,
                              client_domain=self.domain,
                              require_membership=self.group_sid,
                              client_helper="gss-spnego-client",
                              server_helper="gss-spnego",
                              server_use_winbind=True)
        self.assertTrue(ret)

        ret = self.run_helper(client_username=self.username,
                              client_password=self.password,
                              client_domain=self.domain,
                              require_membership=self.bad_group_sid,
                              client_helper="gss-spnego-client",
                              server_helper="gss-spnego",
                              server_use_winbind=True)
        self.assertFalse(ret)

    def test_plaintext_with_membership(self):
        """ ntlm_auth plaintext authentication with require-membership-of """

        proc = Popen([self.ntlm_auth_path,
                      "--require-membership-of", self.group_sid,
                      "--helper-protocol", "squid-2.5-basic"],
                      stdout=PIPE, stdin=PIPE, stderr=PIPE)
        creds = "%s%s%s %s\n" % (self.domain, self.winbind_separator,
                                 self.username,
                                 self.password)
        (out, err) = proc.communicate(input=creds.encode('utf-8'))
        self.assertEqual(proc.returncode, 0)
        self.assertTrue(out.startswith(b"OK\n"))

        # Check membership failure
        proc = Popen([self.ntlm_auth_path,
                      "--require-membership-of", self.bad_group_sid,
                      "--helper-protocol", "squid-2.5-basic"],
                      stdout=PIPE, stdin=PIPE, stderr=PIPE)
        creds = "%s%s%s %s\n" % (self.domain,
                                 self.winbind_separator,
                                 self.username,
                                 self.password)
        (out, err) = proc.communicate(input=creds.encode('utf-8'))
        self.assertEqual(proc.returncode, 0)
        self.assertTrue(out.startswith(b"ERR\n"))

    def test_ntlm_server_1_with_fixed_password(self):
        """ ntlm_auth ntlm-server-1 with fixed password """

        ntlm_cmds = [
            "LANMAN-Challenge: 0123456789abcdef",
            "NT-Response: 25a98c1c31e81847466b29b2df4680f39958fb8c213a9cc6",
            "NT-Domain: TEST",
            "Username: testuser",
            "Request-User-Session-Key: Yes",
            ".\n" ]

        proc = Popen([self.ntlm_auth_path,
                      "--password", "SecREt01",
                      "--helper-protocol", "ntlm-server-1"],
                      stdout=PIPE, stdin=PIPE, stderr=PIPE)
        buf = "\n".join(ntlm_cmds)
        (out, err) = proc.communicate(input=buf.encode('utf-8'))
        self.assertEqual(proc.returncode, 0)

        lines = out.split(b"\n")

        self.assertEqual(len(lines), 4)
        self.assertEqual(lines[0], b"Authenticated: Yes")
        self.assertEqual(
            lines[1], b"User-Session-Key: 3F373EA8E4AF954F14FAA506F8EEBDC4")
        self.assertEqual(lines[2], b".")
        self.assertEqual(lines[3], b"")

        # Break the password with a leading A on the challenge
        ntlm_cmds[0] = "LANMAN-Challenge: A123456789abcdef"

        proc = Popen([self.ntlm_auth_path,
                      "--password", "SecREt01",
                      "--helper-protocol", "ntlm-server-1"],
                      stdout=PIPE, stdin=PIPE, stderr=PIPE)
        buf = "\n".join(ntlm_cmds)
        (out, err) = proc.communicate(input=buf.encode('utf-8'))
        self.assertEqual(proc.returncode, 0)

        lines = out.split(b"\n")
        self.assertEqual(len(lines), 5)
        self.assertEqual(lines[0], b"Authenticated: No")

    def test_ntlm_server_1_with_plaintext_winbind(self):
        """ ntlm_auth ntlm-server-1 with plaintext password against winbind """

        ntlm_cmds = [
            "Password: %s" % self.password,
            "NT-Domain: %s" % self.domain,
            "Username: %s" % self.username,
            "Request-User-Session-Key: Yes",
            ".\n" ]

        proc = Popen([self.ntlm_auth_path,
                      "--require-membership-of", self.group_sid,
                      "--helper-protocol", "ntlm-server-1"],
                      stdout=PIPE, stdin=PIPE, stderr=PIPE)
        buf = "\n".join(ntlm_cmds)
        (out, err) = proc.communicate(input=buf.encode('utf-8'))
        self.assertEqual(proc.returncode, 0)

        lines = out.split(b"\n")

        self.assertEqual(len(lines), 3)
        self.assertEqual(lines[0], b"Authenticated: Yes")
        self.assertEqual(lines[1], b".")
        self.assertEqual(lines[2], b"")

        # Check membership failure

        proc = Popen([self.ntlm_auth_path,
                      "--require-membership-of", self.bad_group_sid,
                      "--helper-protocol", "ntlm-server-1"],
                      stdout=PIPE, stdin=PIPE, stderr=PIPE)
        buf = "\n".join(ntlm_cmds)
        (out, err) = proc.communicate(input=buf.encode('utf-8'))
        self.assertEqual(proc.returncode, 0)

        lines = out.split(b"\n")

        self.assertEqual(len(lines), 3)
        self.assertEqual(lines[0], b"Authenticated: No")
        self.assertEqual(lines[1], b".")
        self.assertEqual(lines[2], b"")

    def test_ntlm_server_1_with_incorrect_password_winbind(self):
        """ ntlm_auth ntlm-server-1 with incorrect fixed password against
        winbind """

        ntlm_cmds = [
            "LANMAN-Challenge: 0123456789abcdef",
            "NT-Response: 25a98c1c31e81847466b29b2df4680f39958fb8c213a9cc6",
            "NT-Domain: %s" % self.domain,
            "Username: %s" % self.username,
            "Request-User-Session-Key: Yes",
            ".\n" ]

        proc = Popen([self.ntlm_auth_path,
                      "--helper-protocol", "ntlm-server-1"],
                      stdout=PIPE, stdin=PIPE, stderr=PIPE)
        buf = "\n".join(ntlm_cmds)
        (out, err) = proc.communicate(input=buf.encode('utf-8'))
        self.assertEqual(proc.returncode, 0)

        lines = out.split(b"\n")

        self.assertEqual(len(lines), 5)
        self.assertEqual(lines[0], b"Authenticated: No")

    def test_diagnostics(self):
        """ ntlm_auth diagnostics """
        cmd_line = [self.ntlm_auth_path,
                    "--username", self.username,
                    "--password", self.password,
                    "--domain", self.domain,
                    "--diagnostics"]
        try:
            self.check_exit_code(cmd_line, 0)
        except BlackboxProcessError as e:
            self.fail(e)

    def test_diagnostics_lm(self):
        """ ntlm_auth diagnostics """
        cmd_line = [self.ntlm_auth_path,
                    "--username", self.username,
                    "--password", self.password,
                    "--domain", self.domain,
                    "--diagnostics",
                    "--request-lm-key"]
        try:
            self.check_exit_code(cmd_line, 0)
        except BlackboxProcessError as e:
            self.fail(e)



# Unix SMB/CIFS implementation.
# Copyright (C) Catalyst.Net Ltd 2023
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

"""Group Key Distribution Service module"""

from enum import Enum
from functools import total_ordering
from typing import Optional, Tuple

from cryptography.hazmat.primitives import hashes

from samba import _glue
from samba.dcerpc import gkdi, misc
from samba.ndr import ndr_pack, ndr_unpack
from samba.nt_time import NtTime, NtTimeDelta


uint64_max: int = 2**64 - 1

L1_KEY_ITERATION: int = _glue.GKDI_L1_KEY_ITERATION
L2_KEY_ITERATION: int = _glue.GKDI_L2_KEY_ITERATION
KEY_CYCLE_DURATION: NtTimeDelta = _glue.GKDI_KEY_CYCLE_DURATION
MAX_CLOCK_SKEW: NtTimeDelta = _glue.GKDI_MAX_CLOCK_SKEW

KEY_LEN_BYTES = 64


class Algorithm(Enum):
    SHA1 = "SHA1"
    SHA256 = "SHA256"
    SHA384 = "SHA384"
    SHA512 = "SHA512"

    def algorithm(self) -> hashes.HashAlgorithm:
        if self is Algorithm.SHA1:
            return hashes.SHA1()

        if self is Algorithm.SHA256:
            return hashes.SHA256()

        if self is Algorithm.SHA384:
            return hashes.SHA384()

        if self is Algorithm.SHA512:
            return hashes.SHA512()

        raise RuntimeError("unknown hash algorithm {self}")

    def __repr__(self) -> str:
        return str(self)

    @staticmethod
    def from_kdf_parameters(kdf_param: Optional[bytes]) -> "Algorithm":
        if not kdf_param:
            return Algorithm.SHA256  # the default used by Windows.

        kdf_parameters = ndr_unpack(gkdi.KdfParameters, kdf_param)
        return Algorithm(kdf_parameters.hash_algorithm)


class GkidType(Enum):
    DEFAULT = object()
    L0_SEED_KEY = object()
    L1_SEED_KEY = object()
    L2_SEED_KEY = object()

    def description(self) -> str:
        if self is GkidType.DEFAULT:
            return "a default GKID"

        if self is GkidType.L0_SEED_KEY:
            return "an L0 seed key"

        if self is GkidType.L1_SEED_KEY:
            return "an L1 seed key"

        if self is GkidType.L2_SEED_KEY:
            return "an L2 seed key"

        raise RuntimeError("unknown GKID type {self}")


class InvalidDerivation(Exception):
    pass


class UndefinedStartTime(Exception):
    pass


@total_ordering
class Gkid:
    # L2 increments every 10 hours. It rolls over after 320 hours (13 days and 8 hours).
    # L1 increments every 320 hours. It rolls over after 10240 hours (426 days and 16 hours).
    # L0 increments every 10240 hours. It rolls over after 43980465111040 hours (five billion years).

    __slots__ = ["_l0_idx", "_l1_idx", "_l2_idx"]

    max_l0_idx = 0x7FFF_FFFF

    def __init__(self, l0_idx: int, l1_idx: int, l2_idx: int) -> None:
        if not -1 <= l0_idx <= Gkid.max_l0_idx:
            raise ValueError(f"L0 index {l0_idx} out of range")

        if not -1 <= l1_idx < L1_KEY_ITERATION:
            raise ValueError(f"L1 index {l1_idx} out of range")

        if not -1 <= l2_idx < L2_KEY_ITERATION:
            raise ValueError(f"L2 index {l2_idx} out of range")

        if l0_idx == -1 and l1_idx != -1:
            raise ValueError("invalid combination of negative and non‐negative indices")

        if l1_idx == -1 and l2_idx != -1:
            raise ValueError("invalid combination of negative and non‐negative indices")

        self._l0_idx = l0_idx
        self._l1_idx = l1_idx
        self._l2_idx = l2_idx

    @property
    def l0_idx(self) -> int:
        return self._l0_idx

    @property
    def l1_idx(self) -> int:
        return self._l1_idx

    @property
    def l2_idx(self) -> int:
        return self._l2_idx

    def gkid_type(self) -> GkidType:
        if self.l0_idx == -1:
            return GkidType.DEFAULT

        if self.l1_idx == -1:
            return GkidType.L0_SEED_KEY

        if self.l2_idx == -1:
            return GkidType.L1_SEED_KEY

        return GkidType.L2_SEED_KEY

    def wrapped_l1_idx(self) -> int:
        if self.l1_idx == -1:
            return L1_KEY_ITERATION

        return self.l1_idx

    def wrapped_l2_idx(self) -> int:
        if self.l2_idx == -1:
            return L2_KEY_ITERATION

        return self.l2_idx

    def derive_l1_seed_key(self) -> "Gkid":
        gkid_type = self.gkid_type()
        if (
            gkid_type is not GkidType.L0_SEED_KEY
            and gkid_type is not GkidType.L1_SEED_KEY
        ):
            raise InvalidDerivation(
                "Invalid attempt to derive an L1 seed key from"
                f" {gkid_type.description()}"
            )

        if self.l1_idx == 0:
            raise InvalidDerivation("No further derivation of L1 seed keys is possible")

        return Gkid(self.l0_idx, self.wrapped_l1_idx() - 1, self.l2_idx)

    def derive_l2_seed_key(self) -> "Gkid":
        gkid_type = self.gkid_type()
        if (
            gkid_type is not GkidType.L1_SEED_KEY
            and gkid_type is not GkidType.L2_SEED_KEY
        ):
            raise InvalidDerivation(
                f"Attempt to derive an L2 seed key from {gkid_type.description()}"
            )

        if self.l2_idx == 0:
            raise InvalidDerivation("No further derivation of L2 seed keys is possible")

        return Gkid(self.l0_idx, self.l1_idx, self.wrapped_l2_idx() - 1)

    def __str__(self) -> str:
        return f"Gkid({self.l0_idx}, {self.l1_idx}, {self.l2_idx})"

    def __repr__(self) -> str:
        cls = type(self)
        return (
            f"{cls.__qualname__}({repr(self.l0_idx)}, {repr(self.l1_idx)},"
            f" {repr(self.l2_idx)})"
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Gkid):
            return NotImplemented

        return (self.l0_idx, self.l1_idx, self.l2_idx) == (
            other.l0_idx,
            other.l1_idx,
            other.l2_idx,
        )

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, Gkid):
            return NotImplemented

        def as_tuple(gkid: Gkid) -> Tuple[int, int, int]:
            l0_idx, l1_idx, l2_idx = gkid.l0_idx, gkid.l1_idx, gkid.l2_idx

            # DEFAULT is considered less than everything else, so that the
            # lexical ordering requirement in [MS-GKDI] 3.1.4.1.3 (GetKey) makes
            # sense.
            if gkid.gkid_type() is not GkidType.DEFAULT:
                # Use the wrapped indices so that L1 seed keys are considered
                # greater than their children L2 seed keys, and L0 seed keys are
                # considered greater than their children L1 seed keys.
                l1_idx = gkid.wrapped_l1_idx()
                l2_idx = gkid.wrapped_l2_idx()

            return l0_idx, l1_idx, l2_idx

        return as_tuple(self) < as_tuple(other)

    def __hash__(self) -> int:
        return hash((self.l0_idx, self.l1_idx, self.l2_idx))

    @staticmethod
    def default() -> "Gkid":
        return Gkid(-1, -1, -1)

    @staticmethod
    def l0_seed_key(l0_idx: int) -> "Gkid":
        return Gkid(l0_idx, -1, -1)

    @staticmethod
    def l1_seed_key(l0_idx: int, l1_idx: int) -> "Gkid":
        return Gkid(l0_idx, l1_idx, -1)

    @staticmethod
    def from_nt_time(nt_time: NtTime) -> "Gkid":
        l0 = nt_time // (L1_KEY_ITERATION * L2_KEY_ITERATION * KEY_CYCLE_DURATION)
        l1 = (
            nt_time
            % (L1_KEY_ITERATION * L2_KEY_ITERATION * KEY_CYCLE_DURATION)
            // (L2_KEY_ITERATION * KEY_CYCLE_DURATION)
        )
        l2 = nt_time % (L2_KEY_ITERATION * KEY_CYCLE_DURATION) // KEY_CYCLE_DURATION

        return Gkid(l0, l1, l2)

    def start_nt_time(self) -> NtTime:
        gkid_type = self.gkid_type()
        if gkid_type is not GkidType.L2_SEED_KEY:
            raise UndefinedStartTime(
                f"{gkid_type.description()} has no defined start time"
            )

        start_time = NtTime(
            (
                self.l0_idx * L1_KEY_ITERATION * L2_KEY_ITERATION
                + self.l1_idx * L2_KEY_ITERATION
                + self.l2_idx
            )
            * KEY_CYCLE_DURATION
        )

        if not 0 <= start_time <= uint64_max:
            raise OverflowError(f"start time {start_time} out of range")

        return start_time

    def previous(self) -> "Gkid":
        return Gkid.from_nt_time(NtTime(self.start_nt_time() - KEY_CYCLE_DURATION))

    def next(self) -> "Gkid":
        return Gkid.from_nt_time(NtTime(self.start_nt_time() + KEY_CYCLE_DURATION))

    @staticmethod
    def from_key_envelope(env: gkdi.KeyEnvelope) -> "Gkid":
        return Gkid(env.l0_index, env.l1_index, env.l2_index)


class SeedKeyPair:
    __slots__ = ["l1_key", "l2_key", "gkid", "hash_algorithm", "root_key_id"]

    def __init__(
        self,
        l1_key: Optional[bytes],
        l2_key: Optional[bytes],
        gkid: Gkid,
        hash_algorithm: Algorithm,
        root_key_id: misc.GUID,
    ) -> None:
        if l1_key is not None and len(l1_key) != KEY_LEN_BYTES:
            raise ValueError(f"L1 key ({repr(l1_key)}) must be {KEY_LEN_BYTES} bytes")
        if l2_key is not None and len(l2_key) != KEY_LEN_BYTES:
            raise ValueError(f"L2 key ({repr(l2_key)}) must be {KEY_LEN_BYTES} bytes")

        self.l1_key = l1_key
        self.l2_key = l2_key
        self.gkid = gkid
        self.hash_algorithm = hash_algorithm
        self.root_key_id = root_key_id

    def __str__(self) -> str:
        l1_key_hex = None if self.l1_key is None else self.l1_key.hex()
        l2_key_hex = None if self.l2_key is None else self.l2_key.hex()

        return (
            f"SeedKeyPair(L1Key({l1_key_hex}), L2Key({l2_key_hex}), {self.gkid},"
            f" {self.root_key_id}, {self.hash_algorithm})"
        )

    def __repr__(self) -> str:
        cls = type(self)
        return (
            f"{cls.__qualname__}({repr(self.l1_key)}, {repr(self.l2_key)},"
            f" {repr(self.gkid)}, {repr(self.hash_algorithm)},"
            f" {repr(self.root_key_id)})"
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SeedKeyPair):
            return NotImplemented

        return (
            self.l1_key,
            self.l2_key,
            self.gkid,
            self.hash_algorithm,
            self.root_key_id,
        ) == (
            other.l1_key,
            other.l2_key,
            other.gkid,
            other.hash_algorithm,
            other.root_key_id,
        )

    def __hash__(self) -> int:
        return hash(
            (
                self.l1_key,
                self.l2_key,
                self.gkid,
                self.hash_algorithm,
                ndr_pack(self.root_key_id),
            )
        )


class GroupKey:
    __slots__ = ["gkid", "key", "hash_algorithm", "root_key_id"]

    def __init__(
        self, key: bytes, gkid: Gkid, hash_algorithm: Algorithm, root_key_id: misc.GUID
    ) -> None:
        if key is not None and len(key) != KEY_LEN_BYTES:
            raise ValueError(f"Key ({repr(key)}) must be {KEY_LEN_BYTES} bytes")

        self.key = key
        self.gkid = gkid
        self.hash_algorithm = hash_algorithm
        self.root_key_id = root_key_id

    def __str__(self) -> str:
        return (
            f"GroupKey(Key({self.key.hex()}), {self.gkid}, {self.hash_algorithm},"
            f" {self.root_key_id})"
        )

    def __repr__(self) -> str:
        cls = type(self)
        return (
            f"{cls.__qualname__}({repr(self.key)}, {repr(self.gkid)},"
            f" {repr(self.hash_algorithm)}, {repr(self.root_key_id)})"
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, GroupKey):
            return NotImplemented

        return (self.key, self.gkid, self.hash_algorithm, self.root_key_id) == (
            other.key,
            other.gkid,
            other.hash_algorithm,
            other.root_key_id,
        )

    def __hash__(self) -> int:
        return hash(
            (self.key, self.gkid, self.hash_algorithm, ndr_pack(self.root_key_id))
        )



# Create forest updates ldif from Github markdown
#
# Each update is converted to an ldif then gets written to a corresponding
# .LDF output file or stored in a dictionary.
#
# Only add updates can generally be applied.
#
# Copyright (C) Andrew Bartlett <abartlet@samba.org> 2017
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Generate LDIF from Github documentation."""

import re
import os
import markdown
import xml.etree.ElementTree as ET
from samba.common import get_string


# Display specifier updates or otherwise (ignored in forest_update.py)
def noop(description, attributes, sd):
    return (None, None, [], None)


# ACE addition updates (ignored in forest_update.py)
def parse_grant(description, attributes, sd):
    # Granting the "CN=Send-As,CN=Extended-Rights" to gMSA accounts.
    if (description.startswith("Granting the ") and
        description.endswith("to gMSA accounts.") and
        (attributes and attributes.lower() == 'n/a') and
        (sd and sd.lower() == 'n/a')):
        return ('modify', extract_dn_or_none(description),
                ['add: appliesTo', 'appliesTo: 7b8b558a-93a5-4af7-adca-c017e67f1057'],
                None)

    return ('modify', None, [], sd if sd.lower() != 'n/a' else None)


# Addition of new objects to the directory (most are applied in forest_update.py)
def parse_add(description, attributes, sd):
    dn = extract_dn(description)
    return ('add', dn, extract_attrib(dn, attributes), sd if sd.lower() != 'n/a' else None)


# Set of a particular attribute (ignored in forest_update.py)
def parse_set(description, attributes, sd):
    return ('modify', extract_dn_or_none(description),
            extract_replace_attrib(attributes),
            sd if sd.lower() != 'n/a' else None)


# Set of a particular ACE (ignored in forest_update.py)
# The general issue is that the list of DNs must be generated dynamically
def parse_ace(description, attributes, sd):

    def extract_dn_ace(text):
        if 'Sam-Domain' in text:
            return ('${DOMAIN_DN}', 'CN=Sam-Domain,${SCHEMA_DN}')
        elif 'Domain-DNS' in text:
            return ('${...}', 'CN=Domain-DNS,${SCHEMA_DN}')

        return None

    return [('modify', extract_dn_ace(description)[0],
             ['replace: nTSecurityDescriptor',
              'nTSecurityDescriptor: ${DOMAIN_SCHEMA_SD}%s' % sd], None),
            ('modify', extract_dn_ace(description)[1],
             ['replace: defaultSecurityDescriptor',
              'defaultSecurityDescriptor: ${OLD_SAMBA_SD}%s' % sd], None)]


# We are really only interested in 'Created' items
operation_map = {
    # modify
    'Granting': parse_grant,
    # add
    'Created': parse_add,
    # modify
    'Set': parse_set,
    # modify
    'Added ACE': parse_ace,
    # modify
    'Updated': parse_set,
    # unknown
    'Call': noop
}


def extract_dn(text):
    """
    Extract a DN from the textual description
    :param text:
    :return: DN in string form
    """
    text = text.replace(' in the Schema partition.', ',${SCHEMA_DN}')
    text = text.replace(' in the Configuration partition.', ',${CONFIG_DN}')
    dn = re.search('([CDO][NCU]=.*?,)*([CDO][NCU]=.*)', text).group(0)

    # This should probably be also fixed upstream
    if dn == 'CN=ad://ext/AuthenticationSilo,CN=Claim Types,CN=Claims Configuration,CN=Services':
        return 'CN=ad://ext/AuthenticationSilo,CN=Claim Types,CN=Claims Configuration,CN=Services,${CONFIG_DN}'

    # Granting the `CN=Send-As,CN=Extended-Rights` to gMSA accounts.
    if dn.endswith(',CN=Extended-Rights to gMSA accounts.'):
        dn = dn.replace(' to gMSA accounts.', '')
        return dn + ",${CONFIG_DN}"

    return dn


def extract_dn_or_none(text):
    """
    Same as above, but returns None if it doesn't work
    :param text:
    :return: DN or None
    """
    try:
        return extract_dn(text)
    except:
        return None


def save_ldif(filename, answers, out_folder):
    """
    Save ldif to disk for each updates
    :param filename: filename use ([OPERATION NUM]-{GUID}.ldif)
    :param answers: array of tuples generated with earlier functions
    :param out_folder: folder to prepend
    """
    path = os.path.join(out_folder, filename)
    with open(path, 'w') as ldif:
        for answer in answers:
            change, dn, attrib, sd = answer
            ldif.write('dn: %s\n' % dn)
            ldif.write('changetype: %s\n' % change)
            if len(attrib) > 0:
                ldif.write('\n'.join(attrib) + '\n')
            if sd is not None:
                ldif.write('nTSecurityDescriptor: D:%s\n' % sd)
            ldif.write('-\n\n')


def save_array(guid, answers, out_dict):
    """
    Save ldif to an output dictionary
    :param guid: GUID to store
    :param answers: array of tuples generated with earlier functions
    :param out_dict: output dictionary
    """
    ldif = ''
    for answer in answers:
        change, dn, attrib, sd = answer
        ldif += 'dn: %s\n' % dn
        ldif += 'changetype: %s\n' % change
        if len(attrib) > 0:
            ldif += '\n'.join(attrib) + '\n'
        if sd is not None:
            ldif += 'nTSecurityDescriptor: D:%s\n' % sd
        ldif += '-\n\n'

    out_dict[guid] = ldif


def extract_attrib(dn, attributes):
    """
    Extract the attributes as an array from the attributes column
    :param dn: parsed from markdown
    :param attributes: from markdown
    :return: attribute array (ldif-type format)
    """
    attrib = [x.lstrip('- ') for x in attributes.split('-   ') if x.lower() != 'n/a' and x != '']
    attrib = [x.replace(': True', ': TRUE') if x.endswith(': True') else x for x in attrib]
    attrib = [x.replace(': False', ': FALSE') if x.endswith(': False') else x for x in attrib]
    # We only have one such value, we may as well skip them all consistently
    attrib = [x for x in attrib if not x.lower().startswith('msds-claimpossiblevalues')]

    return attrib


def extract_replace_attrib(attributes):
    """
    Extract the attributes as an array from the attributes column
    (for replace)
    :param attributes: from markdown
    :return: attribute array (ldif-type format)
    """
    lines = [x.lstrip('- ') for x in attributes.split('-   ') if x.lower() != 'n/a' and x != '']
    lines = [('replace: %s' % line.split(':')[0], line) for line in lines]
    lines = [line for pair in lines for line in pair]
    return lines


def innertext(tag):
    return (tag.text or '') + \
        ''.join(innertext(e) for e in tag) + \
        (tag.tail or '')


def read_ms_markdown(in_file, out_folder=None, out_dict=None):
    """
    Read Github documentation to produce forest wide updates
    :param in_file: Forest-Wide-Updates.md
    :param out_folder: output folder
    :param out_dict: output dictionary
    """

    with open(in_file) as update_file:
        # There is a hidden ClaimPossibleValues in this md file
        content = update_file.read()

        content = re.sub(r'<p>',
                         '<br />',
                         content)
        content = re.sub(r'CN=\\<forest root domain',
                         'CN=<forest root domain',
                         content)

        content = re.sub(r'CN=<forest root domain.*?>',
                         '${FOREST_ROOT_DOMAIN}',
                         content)

        html = markdown.markdown(content,
                                 output_format='xhtml')

    html = html.replace('CN=Schema,%ws', '${SCHEMA_DN}')

    tree = ET.fromstring('<root>' + html + '</root>')

    for node in tree:
        if not node.text:
            continue
        updates = None
        if node.text.startswith('|Operation'):
            # Strip first and last |
            updates = [x[1:len(x) - 1].split('|') for x in
                       get_string(ET.tostring(node, method='text')).splitlines()]
        elif node.text.startswith('| Operation'):
            # Strip first and last |
            updates = [x[2:len(x) - 2].split(' | ') for x in
                       get_string(ET.tostring(node, method='text')).splitlines()]
        if updates:
            for update in updates[2:]:
                output = re.match(r'Operation (\d+): {(.*)}', update[0])
                if not output:
                    raise Exception(update)

                guid = output.group(2)
                filename = "%s-{%s}.ldif" % (output.group(1).zfill(4), guid)

                found = False

                if update[3].startswith('Created') or update[1].startswith('Added ACE'):
                    # Trigger the security descriptor code
                    # Reduce info to just the security descriptor
                    update[3] = update[3].split(':')[-1]

                    result = parse_ace(update[1], update[2], update[3])

                    if filename and out_folder is not None:
                        save_ldif(filename, result, out_folder)
                    else:
                        save_array(guid, result, out_dict)

                    continue

                for operation in operation_map:
                    if update[1].startswith(operation):
                        found = True

                        result = operation_map[operation](update[1], update[2], update[3])

                        if filename and out_folder is not None:
                            save_ldif(filename, [result], out_folder)
                        else:
                            save_array(guid, [result], out_dict)

                        break

                if not found:
                    raise Exception(update)

            # print ET.tostring(node, method='text')


if __name__ == '__main__':
    import sys

    out_folder = ''

    if len(sys.argv) == 0:
        print("Usage: %s <Forest-Wide-Updates.md> [<output folder>]" % (sys.argv[0]), file=sys.stderr)
        sys.exit(1)

    in_file = sys.argv[1]
    if len(sys.argv) > 2:
        out_folder = sys.argv[2]

    read_ms_markdown(in_file, out_folder)



# Unix SMB/CIFS implementation.
# Copyright (C) Kai Blin  <kai@samba.org> 2018
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
from samba import credentials
from samba.dcerpc import dns
from samba.tests.subunitrun import SubunitOptions, TestProgram
from samba.tests.dns_base import DNSTest
import samba.getopt as options
import optparse

parser = optparse.OptionParser("dns_invalid.py <server ip> [options]")
sambaopts = options.SambaOptions(parser)
parser.add_option_group(sambaopts)

# This timeout only has relevance when testing against Windows
# Format errors tend to return patchy responses, so a timeout is needed.
parser.add_option("--timeout", type="int", dest="timeout",
                  help="Specify timeout for DNS requests")

# use command line creds if available
credopts = options.CredentialsOptions(parser)
parser.add_option_group(credopts)
subunitopts = SubunitOptions(parser)
parser.add_option_group(subunitopts)

opts, args = parser.parse_args()

lp = sambaopts.get_loadparm()
creds = credopts.get_credentials(lp)

timeout = opts.timeout

if len(args) < 1:
    parser.print_usage()
    sys.exit(1)

server_ip = args[0]
creds.set_krb_forwardable(credentials.NO_KRB_FORWARDABLE)


class TestBrokenQueries(DNSTest):
    def setUp(self):
        super().setUp()
        global server, server_ip, lp, creds, timeout
        self.server_ip = server_ip
        self.lp = lp
        self.creds = creds
        self.timeout = timeout

    def test_invalid_chars_in_name(self):
        """Check the server refuses invalid characters in the query name"""
        p = self.make_name_packet(dns.DNS_OPCODE_QUERY)
        questions = []

        name = "\x10\x11\x05\xa8.%s" % self.get_dns_domain()
        q = self.make_name_question(name, dns.DNS_QTYPE_A, dns.DNS_QCLASS_IN)
        print("asking for %s" % (q.name))
        questions.append(q)

        self.finish_name_packet(p, questions)
        (response, response_packet) = self.dns_transaction_udp(p, host=server_ip)
        self.assert_dns_rcode_equals(response, dns.DNS_RCODE_NXDOMAIN)


TestProgram(module=__name__, opts=subunitopts)



# Unix SMB/CIFS implementation.
# Copyright (C) Douglas Bagnall <douglas.bagnall@catalyst.net.nz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import tarfile
from pathlib import Path
from tarfile import ExtractError, TarFile as UnsafeTarFile


class TarFile(UnsafeTarFile):
    """This TarFile implementation is trying to ameliorate CVE-2007-4559,
    where tarfile.TarFiles can step outside of the target directory
    using '../../'.
    """

    try:
        # In 3.8.18 (the last 3.8) and post 2023-08-22 versions of
        # 3.9+ (including all of 3.12 and greater), Python's standard
        # tarfile module uses the extraction_filter method for
        # preventing path traversal. See:
        #
        # https://docs.python.org/3/library/tarfile.html#tarfile.TarFile.extraction_filter
        # https://peps.python.org/pep-0706/
        #
        # The default filter is 'fully_trusted', which allows
        # extraction outside the directory, but in Python 3.14 the
        # default is expected to change to the stricter 'data' filter.
        # The 'data' filter preserves no permissions so we select the
        # intermediate 'tar' filter here which prevents escape but
        # preserves permissions.
        #
        # When we no longer support versions less than 3.8 or 3.9, we
        # should remove this whole try...except and just have this
        # next line as the whole class body:
        extraction_filter = staticmethod(tarfile.tar_filter)
    except AttributeError:
        def extract(self, member, path="", set_attrs=True, *,
                    numeric_owner=False):
            self._safetarfile_check()
            super().extract(member, path, set_attrs=set_attrs,
                            numeric_owner=numeric_owner)

        def extractall(self, path, members=None, *, numeric_owner=False):
            self._safetarfile_check()
            super().extractall(path, members,
                               numeric_owner=numeric_owner)

        def _safetarfile_check(self):
            for tarinfo in self.__iter__():
                if self._is_traversal_attempt(tarinfo=tarinfo):
                    raise ExtractError(
                        "Attempted directory traversal for "
                        f"member: {tarinfo.name}")
                if self._is_unsafe_symlink(tarinfo=tarinfo):
                    raise ExtractError(
                        "Attempted directory traversal via symlink for "
                        f"member: {tarinfo.linkname}")
                if self._is_unsafe_link(tarinfo=tarinfo):
                    raise ExtractError(
                        "Attempted directory traversal via link for "
                        f"member: {tarinfo.linkname}")

        def _resolve_path(self, path):
            return os.path.realpath(os.path.abspath(path))

        def _is_path_in_dir(self, path, basedir):
            return self._resolve_path(os.path.join(basedir,
                                      path)).startswith(basedir)

        def _is_traversal_attempt(self, tarinfo):
            if (tarinfo.name.startswith(os.sep)
               or ".." + os.sep in tarinfo.name):
                return True
            return False

        def _is_unsafe_symlink(self, tarinfo):
            if tarinfo.issym():
                symlink_file = Path(
                    os.path.normpath(os.path.join(os.getcwd(),
                                     tarinfo.linkname)))
                if not self._is_path_in_dir(symlink_file, os.getcwd()):
                    return True
            return False

        def _is_unsafe_link(self, tarinfo):
            if tarinfo.islnk():
                link_file = Path(
                    os.path.normpath(os.path.join(os.getcwd(),
                                                  tarinfo.linkname)))
                if not self._is_path_in_dir(link_file, os.getcwd()):
                    return True
            return False


open = TarFile.open



#!/usr/bin/env python3
# Unix SMB/CIFS implementation.
#
# Copyright © 2024 Stefan Metzmacher <metze@samba.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import os

sys.path.insert(0, "bin/python")
os.environ["PYTHONUNBUFFERED"] = "1"

import json

import samba.tests
from samba.credentials import Credentials
from samba.ndr import ndr_print
from samba.dcerpc import witness
from samba.tests import DynamicTestCase, BlackboxTestCase
from samba.common import get_string
from samba import werror, WERRORError

@DynamicTestCase
class RpcdWitnessSambaTests(BlackboxTestCase):
    @classmethod
    def setUpDynamicTestCases(cls):
        cls.num_nodes = int(samba.tests.env_get_var_value('NUM_NODES'))

        def _define_tests(idx1, idx2, ndr64=False):
            cls._define_GetInterfaceList_test(idx1, idx2, ndr64)
            if idx1 == 0 and idx2 != -1:
                cls._define_ResourceChangeCTDB_tests(idx1, idx2, ndr64)

        for idx1 in range(0, cls.num_nodes):
            _define_tests(idx1, -1, ndr64=False)
            _define_tests(idx1, -1, ndr64=True)
            for idx2 in range(0, cls.num_nodes):
                _define_tests(idx1, idx2, ndr64=False)
                _define_tests(idx1, idx2, ndr64=True)

    def setUp(self):
        super().setUp()

        # ctdb/tests/local_daemons.sh doesn't like CTDB_SOCKET to be set already
        # and it doesn't need CTDB_BASE, so we stash them away
        self.saved_CTDB_SOCKET = samba.tests.env_get_var_value('CTDB_SOCKET',
                                                               allow_missing=True)
        if self.saved_CTDB_SOCKET is not None:
            del os.environ["CTDB_SOCKET"]
        self.saved_CTDB_BASE = samba.tests.env_get_var_value('CTDB_BASE',
                                                             allow_missing=True)
        if self.saved_CTDB_BASE is not None:
            del os.environ["CTDB_BASE"]

        self.disabled_idx = -1

        # set this to True in order to get verbose output
        self.verbose = False

        self.ctdb_prefix = samba.tests.env_get_var_value('CTDB_PREFIX')

        self.cluster_share = samba.tests.env_get_var_value('CLUSTER_SHARE')

        self.lp = self.get_loadparm(s3=True)
        self.remote_domain = samba.tests.env_get_var_value('DOMAIN')
        self.remote_user = samba.tests.env_get_var_value('USERNAME')
        self.remote_password = samba.tests.env_get_var_value('PASSWORD')
        self.remote_creds = Credentials()
        self.remote_creds.guess(self.lp)
        self.remote_creds.set_username(self.remote_user)
        self.remote_creds.set_domain(self.remote_domain)
        self.remote_creds.set_password(self.remote_password)

        self.server_hostname = samba.tests.env_get_var_value('SERVER_HOSTNAME')
        self.interface_group_name = samba.tests.env_get_var_value('INTERFACE_GROUP_NAME')

        common_binding_args = "spnego,sign,target_hostname=%s" % (
            self.server_hostname)
        if self.verbose:
            common_binding_args += ",print"

        common_binding_args32 = common_binding_args
        common_binding_args64 = common_binding_args + ",ndr64"

        self.nodes = []
        for node_idx in range(0, self.num_nodes):
            node = {}

            name_var = 'CTDB_SERVER_NAME_NODE%u' % node_idx
            node["name"] = samba.tests.env_get_var_value(name_var)

            ip_var = 'CTDB_IFACE_IP_NODE%u' % node_idx
            node["ip"] = samba.tests.env_get_var_value(ip_var)

            node["binding_string32"] = "ncacn_ip_tcp:%s[%s]" % (
                    node["ip"], common_binding_args32)
            node["binding_string64"] = "ncacn_ip_tcp:%s[%s]" % (
                    node["ip"], common_binding_args64)
            self.nodes.append(node)

        self.all_registrations = None

    def tearDown(self):
        self.destroy_all_registrations()

        if self.disabled_idx != -1:
            self.enable_node(self.disabled_idx)

        if self.saved_CTDB_SOCKET is not None:
            os.environ["CTDB_SOCKET"] = self.saved_CTDB_SOCKET
            self.saved_CTDB_SOCKET = None
        if self.saved_CTDB_BASE is not None:
            os.environ["CTDB_BASE"] = self.saved_CTDB_BASE
            self.saved_CTDB_BASE = None

        super().tearDown()

    def call_onnode(self, nodes, cmd):
        COMMAND = "ctdb/tests/local_daemons.sh"

        argv = "%s '%s' onnode %s '%s'" % (COMMAND, self.ctdb_prefix, nodes, cmd)

        try:
            if self.verbose:
                print("Calling: %s" % argv)
            out = self.check_output(argv)
        except samba.tests.BlackboxProcessError as e:
            self.fail("Error calling [%s]: %s" % (argv, e))

        out_str = get_string(out)
        return out_str

    def dump_ctdb_status_all(self):
        for node_idx in range(0, self.num_nodes):
            print("%s" % self.call_onnode(str(node_idx), "ctdb status"))

    def disable_node(self, node_idx, dump_status=False):
        if dump_status:
            self.dump_ctdb_status_all()

        self.assertEqual(self.disabled_idx, -1)
        self.call_onnode(str(node_idx), "ctdb disable")
        self.disabled_idx = node_idx

        if dump_status:
            self.dump_ctdb_status_all()

    def enable_node(self, node_idx, dump_status=False):
        if dump_status:
            self.dump_ctdb_status_all()

        self.assertEqual(self.disabled_idx, node_idx)
        self.call_onnode(str(node_idx), "ctdb enable")
        self.disabled_idx = -1

        if dump_status:
            self.dump_ctdb_status_all()

    def call_net_witness_subcmd(self, subcmd,
                                as_json=False,
                                apply_to_all=False,
                                registration=None,
                                net_name=None,
                                share_name=None,
                                ip_address=None,
                                client_computer=None,
                                new_ip=None,
                                new_node=None,
                                forced_response=None):
        COMMAND = "UID_WRAPPER_ROOT=1 bin/net witness"

        argv = "%s %s" % (COMMAND, subcmd)
        if as_json:
            argv += " --json"

        if apply_to_all:
            argv += " --witness-apply-to-all"

        if registration is not None:
            argv += " --witness-registration='%s'" % (
                    registration.uuid)

        if net_name is not None:
            argv += " --witness-net-name='%s'" % (net_name)

        if share_name is not None:
            argv += " --witness-share-name='%s'" % (share_name)

        if ip_address is not None:
            argv += " --witness-ip-address='%s'" % (ip_address)

        if client_computer is not None:
            argv += " --witness-client-computer-name='%s'" % (client_computer)

        if new_ip is not None:
            argv += " --witness-new-ip='%s'" % (new_ip)

        if new_node is not None:
            argv += " --witness-new-node='%s'" % (new_node)

        if forced_response:
            argv += " --witness-forced-response='%s'" % (forced_response)

        try:
            if self.verbose:
                print("Calling: %s" % argv)
            out = self.check_output(argv)
        except samba.tests.BlackboxProcessError as e:
            self.fail("Error calling [%s]: %s" % (argv, e))

        out_str = get_string(out)
        if not as_json:
            return out_str

        json_out = json.loads(out_str)
        return json_out

    @classmethod
    def _define_GetInterfaceList_test(cls, conn_idx, disable_idx, ndr64=False):
        if disable_idx != -1:
            disable_name = "%u_disabled" % disable_idx
        else:
            disable_name = "all_enabled"

        if ndr64:
            ndr_name = "NDR64"
        else:
            ndr_name = "NDR32"

        name = "Node%u_%s_%s" % (conn_idx, disable_name, ndr_name)
        args = {
            'conn_idx': conn_idx,
            'disable_idx': disable_idx,
            'ndr64': ndr64,
        }
        cls.generate_dynamic_test('test_GetInterfaceList', name, args)

    def _test_GetInterfaceList_with_args(self, args):
        conn_idx = args.pop('conn_idx')
        disable_idx = args.pop('disable_idx')
        ndr64 = args.pop('ndr64')
        self.assertEqual(len(args.keys()), 0)

        conn_node = self.nodes[conn_idx]
        if ndr64:
            binding_string = conn_node["binding_string64"]
        else:
            binding_string = conn_node["binding_string32"]

        if disable_idx != -1:
            self.disable_node(disable_idx)

        conn = witness.witness(binding_string, self.lp, self.remote_creds)
        interface_list = conn.GetInterfaceList()

        if disable_idx != -1:
            self.enable_node(disable_idx)

        self.assertIsNotNone(interface_list)
        self.assertEqual(interface_list.num_interfaces, len(self.nodes))
        for idx in range(0, interface_list.num_interfaces):
            iface = interface_list.interfaces[idx]
            node = self.nodes[idx]

            expected_flags = 0
            expected_flags |= witness.WITNESS_INFO_IPv4_VALID
            if conn_idx != idx:
                expected_flags |= witness.WITNESS_INFO_WITNESS_IF

            if disable_idx == idx:
                expected_state = witness.WITNESS_STATE_UNAVAILABLE
            else:
                expected_state = witness.WITNESS_STATE_AVAILABLE

            self.assertIsNotNone(iface.group_name)
            self.assertEqual(iface.group_name.lower(),
                             self.interface_group_name.lower())

            self.assertEqual(iface.version, witness.WITNESS_V2)
            self.assertEqual(iface.state, expected_state)

            self.assertIsNotNone(iface.ipv4)
            self.assertEqual(iface.ipv4, node["ip"])

            self.assertIsNotNone(iface.ipv6)
            self.assertEqual(iface.ipv6,
                    "0000:0000:0000:0000:0000:0000:0000:0000")

            self.assertEqual(iface.flags, expected_flags)

    def assertResourceChanges(self, response, expected_resource_changes):
        self.assertIsNotNone(response)
        self.assertEqual(response.type,
                witness.WITNESS_NOTIFY_RESOURCE_CHANGE)
        self.assertEqual(response.num, len(expected_resource_changes))
        self.assertEqual(len(response.messages), len(expected_resource_changes))
        for ri in range(0, len(expected_resource_changes)):
            expected_resource_change = expected_resource_changes[ri]
            resource_change = response.messages[ri]
            self.assertIsNotNone(resource_change)

            expected_type = witness.WITNESS_RESOURCE_STATE_UNAVAILABLE
            expected_type = expected_resource_change.get('type', expected_type)

            expected_name = expected_resource_change.get('name')

            self.assertEqual(resource_change.type, expected_type)
            self.assertIsNotNone(resource_change.name)
            self.assertEqual(resource_change.name, expected_name)

    def assertResourceChange(self, response, expected_type, expected_name):
        expected_resource_change = {
            'type': expected_type,
            'name': expected_name,
        }
        expected_resource_changes = [expected_resource_change]
        self.assertResourceChanges(response, expected_resource_changes)

    def assertGenericIpLists(self, response, expected_type, expected_ip_lists):
        self.assertIsNotNone(response)
        self.assertEqual(response.type, expected_type)
        self.assertEqual(response.num, len(expected_ip_lists))
        self.assertEqual(len(response.messages), len(expected_ip_lists))
        for li in range(0, len(expected_ip_lists)):

            expected_ip_list = expected_ip_lists[li]
            ip_list = response.messages[li]
            self.assertIsNotNone(ip_list)
            self.assertEqual(ip_list.num, len(expected_ip_list))

            for i in range(0, len(expected_ip_list)):
                ip_info = ip_list.addr[i]

                expected_flags = 0
                expected_flags |= witness.WITNESS_IPADDR_V4
                expected_flags |= witness.WITNESS_IPADDR_ONLINE
                expected_flags = expected_ip_list[i].get('flags', expected_flags)

                expected_ipv4 = '0.0.0.0'
                expected_ipv4 = expected_ip_list[i].get('ipv4', expected_ipv4)

                expected_ipv6 = '0000:0000:0000:0000:0000:0000:0000:0000'
                expected_ipv6 = expected_ip_list[i].get('ipv6', expected_ipv6)

                self.assertEqual(ip_info.flags, expected_flags)

                self.assertIsNotNone(ip_info.ipv4)
                self.assertEqual(ip_info.ipv4, expected_ipv4)

                self.assertIsNotNone(ip_info.ipv6)
                self.assertEqual(ip_info.ipv6, expected_ipv6)

    @classmethod
    def _define_ResourceChangeCTDB_tests(cls, conn_idx, monitor_idx, ndr64=False):
        if ndr64:
            ndr_name = "NDR64"
        else:
            ndr_name = "NDR32"

        name_suffix = "WNode%u_RNode%u_%s" % (conn_idx, monitor_idx, ndr_name)
        base_args = {
            'conn_idx': conn_idx,
            'monitor_idx': monitor_idx,
            'ndr64': ndr64,
        }

        name = "v1_disabled_after_%s" % name_suffix
        args = base_args.copy()
        args['reg_v1'] = True
        args['disable_after_reg'] = True
        args['explicit_unregister'] = False
        cls.generate_dynamic_test('test_ResourceChangeCTDB', name, args)

        name = "v1_disabled_after_enabled_after_%s" % name_suffix
        args = base_args.copy()
        args['reg_v1'] = True
        args['disable_after_reg'] = True
        args['enable_after_reg'] = True
        args['explicit_unregister'] = False
        cls.generate_dynamic_test('test_ResourceChangeCTDB', name, args)

        name = "v2_disabled_before_enable_after_%s" % name_suffix
        args = base_args.copy()
        args['disable_before_reg'] = True
        args['enable_after_reg'] = True
        args['wait_for_timeout'] = True
        args['timeout'] = 6
        cls.generate_dynamic_test('test_ResourceChangeCTDB', name, args)

        name = "v2_disabled_after_%s" % name_suffix
        args = base_args.copy()
        args['disable_after_reg'] = True
        args['wait_for_not_found'] = True
        args['explicit_unregister'] = False
        cls.generate_dynamic_test('test_ResourceChangeCTDB', name, args)

        name = "v2_disabled_after_enabled_after_%s" % name_suffix
        args = base_args.copy()
        args['disable_after_reg'] = True
        args['enable_after_reg'] = True
        args['wait_for_not_found'] = True
        args['explicit_unregister'] = False
        cls.generate_dynamic_test('test_ResourceChangeCTDB', name, args)

        name = "share_v2_disabled_before_enable_after_%s" % name_suffix
        args = base_args.copy()
        args['share_reg'] = True
        args['disable_before_reg'] = True
        args['enable_after_reg'] = True
        cls.generate_dynamic_test('test_ResourceChangeCTDB', name, args)

        name = "share_v2_disabled_after_%s" % name_suffix
        args = base_args.copy()
        args['share_reg'] = True
        args['disable_after_reg'] = True
        args['explicit_unregister'] = False
        cls.generate_dynamic_test('test_ResourceChangeCTDB', name, args)

        name = "share_v2_disabled_after_enabled_after_%s" % name_suffix
        args = base_args.copy()
        args['share_reg'] = True
        args['disable_after_reg'] = True
        args['enable_after_reg'] = True
        args['explicit_unregister'] = False
        cls.generate_dynamic_test('test_ResourceChangeCTDB', name, args)

    def _test_ResourceChangeCTDB_with_args(self, args):
        conn_idx = args.pop('conn_idx')
        monitor_idx = args.pop('monitor_idx')
        ndr64 = args.pop('ndr64')
        timeout = int(args.pop('timeout', 15))
        reg_v1 = args.pop('reg_v1', False)
        share_reg = args.pop('share_reg', False)
        disable_before_reg = args.pop('disable_before_reg', False)
        disable_after_reg = args.pop('disable_after_reg', False)
        enable_after_reg = args.pop('enable_after_reg', False)
        explicit_unregister = args.pop('explicit_unregister', True)
        wait_for_not_found = args.pop('wait_for_not_found', False)
        wait_for_timeout = args.pop('wait_for_timeout', False)
        self.assertEqual(len(args.keys()), 0)

        conn_node = self.nodes[conn_idx]
        if ndr64:
            binding_string = conn_node["binding_string64"]
        else:
            binding_string = conn_node["binding_string32"]
        monitor_node = self.nodes[monitor_idx]

        computer_name = "test-rpcd-witness-samba-only-client-computer"

        conn = witness.witness(binding_string, self.lp, self.remote_creds)

        if disable_before_reg:
            self.assertFalse(disable_after_reg)
            self.disable_node(monitor_idx)

        if reg_v1:
            self.assertFalse(wait_for_timeout)
            self.assertFalse(share_reg)

            reg_context = conn.Register(witness.WITNESS_V1,
                                        self.server_hostname,
                                        monitor_node["ip"],
                                        computer_name)
        else:
            if share_reg:
                share_name = self.cluster_share
            else:
                share_name = None

            reg_context = conn.RegisterEx(witness.WITNESS_V2,
                                          self.server_hostname,
                                          share_name,
                                          monitor_node["ip"],
                                          computer_name,
                                          witness.WITNESS_REGISTER_NONE,
                                          timeout)

        if disable_after_reg:
            self.assertFalse(disable_before_reg)
            self.disable_node(monitor_idx)

        if enable_after_reg:
            self.enable_node(monitor_idx)

        if disable_after_reg:
            response_unavailable = conn.AsyncNotify(reg_context)
            self.assertResourceChange(response_unavailable,
                                      witness.WITNESS_RESOURCE_STATE_UNAVAILABLE,
                                      monitor_node["ip"])

        if enable_after_reg:
            response_available = conn.AsyncNotify(reg_context)
            self.assertResourceChange(response_available,
                                      witness.WITNESS_RESOURCE_STATE_AVAILABLE,
                                      monitor_node["ip"])

        if wait_for_timeout:
            self.assertFalse(wait_for_not_found)
            self.assertFalse(disable_after_reg)
            try:
                _ = conn.AsyncNotify(reg_context)
                self.fail()
            except WERRORError as e:
                (num, string) = e.args
                if num != werror.WERR_TIMEOUT:
                    raise

        if wait_for_not_found:
            self.assertFalse(wait_for_timeout)
            self.assertTrue(disable_after_reg)
            self.assertFalse(explicit_unregister)
            try:
                _ = conn.AsyncNotify(reg_context)
                self.fail()
            except WERRORError as e:
                (num, string) = e.args
                if num != werror.WERR_NOT_FOUND:
                    raise

        if not explicit_unregister:
            return

        conn.UnRegister(reg_context)

        try:
            _ = conn.AsyncNotify(reg_context)
            self.fail()
        except WERRORError as e:
            (num, string) = e.args
            if num != werror.WERR_NOT_FOUND:
                raise

        try:
            conn.UnRegister(reg_context)
            self.fail()
        except WERRORError as e:
            (num, string) = e.args
            if num != werror.WERR_NOT_FOUND:
                raise

    def prepare_all_registrations(self):
        self.assertIsNone(self.all_registrations)

        regs = []
        for node_idx in range(0, self.num_nodes):
            node = self.nodes[node_idx]
            for ndr64 in [False, True]:
                if ndr64:
                    binding_string = node["binding_string64"]
                    ndr_name = "NDR64"
                else:
                    binding_string = node["binding_string32"]
                    ndr_name = "NDR32"

                conn = witness.witness(binding_string, self.lp, self.remote_creds)
                conn_ip = node["ip"]

                net_name = self.server_hostname
                ip_address = node["ip"]
                share_name = self.cluster_share
                computer_name = "test-net-witness-list-%s-%s" % (
                        node_idx, ndr_name)
                flags = witness.WITNESS_REGISTER_NONE
                timeout = 15

                reg_version = witness.WITNESS_V1
                reg = {
                    'node_idx': node_idx,
                    'ndr64': ndr64,
                    'binding_string': binding_string,
                    'conn_ip': conn_ip,
                    'version': reg_version,
                    'net_name': net_name,
                    'share_name': None,
                    'ip_address': ip_address,
                    'computer_name': computer_name,
                    'flags': 0,
                    'timeout': 0,
                    'conn': conn,
                    'context': None,
                }
                regs.append(reg)

                reg_version = witness.WITNESS_V2
                reg = {
                    'node_idx': node_idx,
                    'ndr64': ndr64,
                    'binding_string': binding_string,
                    'conn_ip': conn_ip,
                    'version': reg_version,
                    'net_name': net_name,
                    'share_name': None,
                    'ip_address': ip_address,
                    'computer_name': computer_name,
                    'flags': flags,
                    'timeout': timeout,
                    'conn': conn,
                    'context': None,
                }
                regs.append(reg)

                reg = {
                    'node_idx': node_idx,
                    'ndr64': ndr64,
                    'binding_string': binding_string,
                    'conn_ip': conn_ip,
                    'version': reg_version,
                    'net_name': net_name,
                    'share_name': share_name,
                    'ip_address': ip_address,
                    'computer_name': computer_name,
                    'flags': flags,
                    'timeout': timeout,
                    'conn': conn,
                    'context': None,
                }
                regs.append(reg)

        self.all_registrations = regs
        return regs

    def close_all_registrations(self):
        self.assertIsNotNone(self.all_registrations)

        for reg in self.all_registrations:
            conn = reg['conn']
            reg_context = reg['context']
            if reg_context is not None:
                conn.UnRegister(reg_context)
                reg_context = None
            reg['context'] = reg_context

    def open_all_registrations(self):
        self.assertIsNotNone(self.all_registrations)

        for reg in self.all_registrations:
            conn = reg['conn']
            reg_context = reg['context']
            self.assertIsNone(reg_context)

            reg_version = reg['version']
            if reg_version == witness.WITNESS_V1:
                reg_context = conn.Register(reg_version,
                                            reg['net_name'],
                                            reg['ip_address'],
                                            reg['computer_name'])
            elif reg_version == witness.WITNESS_V2:
                reg_context = conn.RegisterEx(reg_version,
                                              reg['net_name'],
                                              reg['share_name'],
                                              reg['ip_address'],
                                              reg['computer_name'],
                                              reg['flags'],
                                              reg['timeout'])
            self.assertIsNotNone(reg_context)
            reg['context'] = reg_context

    def destroy_all_registrations(self):
        if self.all_registrations is None:
            return

        for reg in self.all_registrations:
            conn = reg['conn']
            reg_context = reg['context']
            if reg_context is not None:
                conn.UnRegister(reg_context)
                reg_context = None
            reg['context'] = reg_context
            conn = None
            reg['conn'] = conn

        self.all_registrations = None

    def assertJsonReg(self, json_reg, reg):
        self.assertEqual(json_reg['version'], "0x%08x" % reg['version'])
        self.assertEqual(json_reg['net_name'], reg['net_name'])
        if reg['share_name']:
            self.assertEqual(json_reg['share_name'], reg['share_name'])
        else:
            self.assertIsNone(json_reg['share_name'])
        self.assertEqual(json_reg['client_computer_name'], reg['computer_name'])

        self.assertIn('flags', json_reg)
        json_flags = json_reg['flags']
        if reg['flags'] & witness.WITNESS_REGISTER_IP_NOTIFICATION:
            expected_ip_notifaction = True
        else:
            expected_ip_notifaction = False
        self.assertEqual(json_flags['WITNESS_REGISTER_IP_NOTIFICATION'],
                         expected_ip_notifaction)
        self.assertEqual(json_flags['int'], reg['flags'])
        self.assertEqual(json_flags['hex'], "0x%08x" % reg['flags'])
        self.assertEqual(len(json_flags.keys()), 3)

        self.assertEqual(json_reg['timeout'], reg['timeout'])

        self.assertIn('context_handle', json_reg)
        json_context = json_reg['context_handle']
        self.assertEqual(json_context['uuid'], str(reg['context'].uuid))
        self.assertEqual(json_context['handle_type'], reg['context'].handle_type)
        self.assertEqual(len(json_context.keys()), 2)

        self.assertIn('server_id', json_reg)
        json_server_id = json_reg['server_id']
        self.assertIn('pid', json_server_id)
        self.assertIn('task_id', json_server_id)
        self.assertEqual(json_server_id['vnn'], reg['node_idx'])
        self.assertIn('unique_id', json_server_id)
        self.assertEqual(len(json_server_id.keys()), 4)

        self.assertIn('auth', json_reg)
        json_auth = json_reg['auth']
        self.assertEqual(json_auth['account_name'], self.remote_user)
        self.assertEqual(json_auth['domain_name'], self.remote_domain)
        self.assertIn('account_sid', json_auth)
        self.assertEqual(len(json_auth.keys()), 3)

        self.assertIn('connection', json_reg)
        json_conn = json_reg['connection']
        self.assertIn('local_address', json_conn)
        self.assertIn(reg['conn_ip'], json_conn['local_address'])
        self.assertIn('remote_address', json_conn)
        self.assertEqual(len(json_conn.keys()), 2)

        self.assertIn('registration_time', json_reg)

        self.assertEqual(len(json_reg.keys()), 12)

    def max_common_prefix(self, strings):
        if len(strings) == 0:
            return ""

        def string_match_len(s1, s2):
            idx = 0
            for i in range(0, min(len(s1), len(s2))):
                c1 = s1[i:i+1]
                c2 = s2[i:i+1]
                if c1 != c2:
                    break
                idx = i
            return idx

        prefix = None
        for s in strings:
            if prefix is None:
                prefix = s
                continue
            l = string_match_len(prefix, s)
            prefix = prefix[0:l+1]

        return prefix

    def check_net_witness_output(self,
                                 cmd,
                                 regs,
                                 apply_to_all=False,
                                 registration_idx=None,
                                 net_name=None,
                                 share_name=None,
                                 ip_address=None,
                                 client_computer=None,
                                 new_ip=None,
                                 new_node=None,
                                 forced_response=None,
                                 expected_msg_type=None,
                                 callback=None):
        self.open_all_registrations()
        if registration_idx is not None:
            registration = regs[registration_idx]['context']
            self.assertIsNotNone(registration)
        else:
            registration = None

        plain_res = self.call_net_witness_subcmd(cmd,
                                                 apply_to_all=apply_to_all,
                                                 registration=registration,
                                                 net_name=net_name,
                                                 share_name=share_name,
                                                 ip_address=ip_address,
                                                 client_computer=client_computer,
                                                 new_ip=new_ip,
                                                 new_node=new_node,
                                                 forced_response=forced_response)
        if self.verbose:
            print("%s" % plain_res)
        plain_lines = plain_res.splitlines()

        num_headlines = 2
        if expected_msg_type:
            num_headlines += 1
        self.assertEqual(len(plain_lines), num_headlines+len(regs))
        if expected_msg_type:
            self.assertIn(expected_msg_type, plain_lines[0])
        plain_lines = plain_lines[num_headlines:]
        self.assertEqual(len(plain_lines), len(regs))

        for reg in regs:
            reg_uuid = reg['context'].uuid

            expected_line = "%-36s " % reg_uuid
            expected_line += "%-20s " % reg['net_name']
            if reg['share_name']:
                expected_share = reg['share_name']
            else:
                expected_share = "''"
            expected_line += "%-15s " % expected_share
            expected_line += "%-20s " % reg['ip_address']
            expected_line += "%s" % reg['computer_name']

            line = None
            for l in plain_lines:
                if not l.startswith(str(reg_uuid)):
                    continue
                self.assertIsNone(line)
                line = l
                self.assertEqual(line, expected_line)
            self.assertIsNotNone(line)

            if callback is not None:
                callback(reg)

        self.close_all_registrations()

        self.open_all_registrations()
        if registration_idx is not None:
            registration = regs[registration_idx]['context']
            self.assertIsNotNone(registration)
        else:
            registration = None

        json_res = self.call_net_witness_subcmd(cmd,
                                                as_json=True,
                                                apply_to_all=apply_to_all,
                                                registration=registration,
                                                net_name=net_name,
                                                share_name=share_name,
                                                ip_address=ip_address,
                                                client_computer=client_computer,
                                                new_ip=new_ip,
                                                new_node=new_node,
                                                forced_response=forced_response)

        num_filters = 0
        if apply_to_all:
            num_filters += 1
        if registration:
            num_filters += 1
        if net_name:
            num_filters += 1
        if share_name:
            num_filters += 1
        if ip_address:
            num_filters += 1
        if client_computer:
            num_filters += 1

        num_toplevel = 2
        if expected_msg_type:
            num_toplevel += 1

        self.assertIn('filters', json_res);
        if expected_msg_type:
            self.assertIn('message', json_res);
        self.assertIn('registrations', json_res);
        self.assertEqual(len(json_res.keys()), num_toplevel)

        json_filters = json_res['filters']
        self.assertEqual(len(json_filters.keys()), num_filters)

        if apply_to_all:
            self.assertTrue(json_filters['--witness-apply-to-all'])

        if registration:
            self.assertEqual(json_filters['--witness-registration'],
                             str(registration.uuid))
        if net_name:
            self.assertEqual(json_filters['--witness-net-name'],
                             net_name)
        if share_name:
            self.assertEqual(json_filters['--witness-share-name'],
                             share_name)
        if ip_address:
            self.assertEqual(json_filters['--witness-ip-address'],
                             ip_address)
        if client_computer:
            self.assertEqual(json_filters['--witness-client-computer-name'],
                             client_computer)
        if expected_msg_type:
            json_message = json_res['message']
            num_sub = 1
            self.assertEqual(json_message['type'], expected_msg_type);

            if new_ip is not None:
                num_sub += 1
                self.assertEqual(json_message['new_ip'], new_ip);
            elif new_node == -1:
                num_sub += 1
                self.assertTrue(json_message['all_nodes'])
            elif new_node is not None:
                num_sub += 1
                self.assertEqual(json_message['new_node'], new_node)
            if forced_response is not None:
                num_sub += 1
                forced_response_json = json.loads(str(forced_response))
                self.assertDictEqual(json_message['json'], forced_response_json)

            self.assertEqual(len(json_message.keys()), num_sub)

        json_regs = json_res['registrations']
        self.assertEqual(len(json_regs.keys()), len(regs))

        for reg in regs:
            reg_uuid = reg['context'].uuid

            self.assertIn(str(reg_uuid), json_regs)
            json_reg = json_regs[str(reg_uuid)]
            self.assertJsonReg(json_reg, reg)

            if callback is not None:
                callback(reg)

        self.close_all_registrations()

    def check_combinations(self, check_func, only_shares=False):
        all_regs = self.prepare_all_registrations()

        share_name_regs = {}
        all_share_name_regs = []
        no_share_name_regs = []
        for reg in all_regs:
            if reg['share_name'] is not None:
                if reg['share_name'] not in share_name_regs:
                    share_name_regs[reg['share_name']] = []
                share_name_regs[reg['share_name']].append(reg)
                all_share_name_regs.append(reg)
            else:
                no_share_name_regs.append(reg)

        if only_shares:
            all_regs = all_share_name_regs
            no_share_name_regs = []

        ip_address_regs = {}
        computer_name_regs = {}
        for reg in all_regs:
            if reg['ip_address'] not in ip_address_regs:
                ip_address_regs[reg['ip_address']] = []
            ip_address_regs[reg['ip_address']].append(reg)

            if reg['computer_name'] not in computer_name_regs:
                computer_name_regs[reg['computer_name']] = []
            computer_name_regs[reg['computer_name']].append(reg)

        all_share_names = '|'.join(share_name_regs.keys())
        common_share_name = self.max_common_prefix(share_name_regs.keys())
        all_ip_addresses = '|'.join(ip_address_regs.keys())
        common_ip_address = self.max_common_prefix(ip_address_regs.keys())
        all_computer_names = '|'.join(computer_name_regs.keys())
        common_computer_name = self.max_common_prefix(computer_name_regs.keys())

        check_func(all_regs,
                   apply_to_all=True)
        check_func(all_regs,
                   net_name=self.server_hostname)
        check_func(all_regs,
                   ip_address=all_ip_addresses)
        check_func(all_regs,
                   client_computer=all_computer_names)
        check_func(all_regs,
                   net_name=self.server_hostname,
                   ip_address=all_ip_addresses,
                   client_computer=all_computer_names)
        check_func(all_regs,
                   net_name='.*',
                   share_name='.*',
                   ip_address='.*',
                   client_computer='.*')
        check_func(all_regs,
                   share_name='^$|%s.*' % common_share_name,
                   ip_address='%s.*' % common_ip_address,
                   client_computer='%s.*' % common_computer_name)
        check_func(all_share_name_regs,
                   share_name=all_share_names)
        check_func(all_share_name_regs,
                   share_name='%s.*' % common_share_name)
        check_func(no_share_name_regs,
                   share_name='^$')

        for share_name in share_name_regs.keys():
            regs = share_name_regs[share_name]
            check_func(regs, share_name=share_name)

        for ip_address in ip_address_regs.keys():
            regs = ip_address_regs[ip_address]
            check_func(regs, ip_address=ip_address)

        for computer_name in computer_name_regs.keys():
            regs = computer_name_regs[computer_name]
            check_func(regs, client_computer=computer_name)

        for reg in all_regs:
            regs = [reg]
            check_func(regs,
                       registration_idx=0)
            check_func(regs,
                       registration_idx=0,
                       net_name=reg['net_name'],
                       share_name=reg['share_name'],
                       ip_address=reg['ip_address'],
                       client_computer=reg['computer_name'])

    def test_net_witness_list(self):
        def check_list(regs,
                       apply_to_all=False,
                       registration_idx=None,
                       net_name=None,
                       share_name=None,
                       ip_address=None,
                       client_computer=None):
            # --witness-apply-to-all is not needed for 'list'
            apply_to_all = None
            return self.check_net_witness_output('list',
                                                 regs,
                                                 apply_to_all=apply_to_all,
                                                 registration_idx=registration_idx,
                                                 net_name=net_name,
                                                 share_name=share_name,
                                                 ip_address=ip_address,
                                                 client_computer=client_computer)

        self.check_combinations(check_list)

    def _test_net_witness_generic_move(self,
                                       move_cmd,
                                       msg_type_prefix,
                                       msg_type):
        def _check_generic_move(regs,
                                apply_to_all=False,
                                registration_idx=None,
                                net_name=None,
                                share_name=None,
                                ip_address=None,
                                client_computer=None,
                                new_ip=None,
                                new_node=None):

            if new_ip:
                expected_msg_type = "%s_IPV4" % msg_type_prefix
            else:
                expected_msg_type = "%s_NODE" % msg_type_prefix

            expected_ip_list = []
            if new_ip:
                ip = { 'ipv4': str(new_ip), }
                expected_ip_list.append(ip)
            if new_node == -1:
                for node_idx in range(0, len(self.nodes)):
                    node = self.nodes[node_idx]
                    ip = { 'ipv4': str(node['ip']), }
                    expected_ip_list.append(ip)
            elif new_node is not None:
                node = self.nodes[new_node]
                ip = { 'ipv4': str(node['ip']), }
                expected_ip_list.append(ip)

            expected_ip_lists = [expected_ip_list]

            def check_generic_move_response(reg):
                conn = reg['conn']
                reg_context = reg['context']
                response = conn.AsyncNotify(reg_context)
                self.assertGenericIpLists(response, msg_type, expected_ip_lists)

            return self.check_net_witness_output(move_cmd,
                                                 regs,
                                                 apply_to_all=apply_to_all,
                                                 registration_idx=registration_idx,
                                                 net_name=net_name,
                                                 share_name=share_name,
                                                 ip_address=ip_address,
                                                 client_computer=client_computer,
                                                 new_ip=new_ip,
                                                 new_node=new_node,
                                                 expected_msg_type=expected_msg_type,
                                                 callback=check_generic_move_response)

        def check_generic_move(regs,
                               apply_to_all=False,
                               registration_idx=None,
                               net_name=None,
                               share_name=None,
                               ip_address=None,
                               client_computer=None):
            _check_generic_move(regs,
                                apply_to_all=apply_to_all,
                                registration_idx=registration_idx,
                                net_name=net_name,
                                share_name=share_name,
                                ip_address=ip_address,
                                client_computer=client_computer,
                                new_node=-1)

            for node_idx in range(0, len(self.nodes)):
                node = self.nodes[node_idx]

                _check_generic_move(regs,
                                    apply_to_all=apply_to_all,
                                    registration_idx=registration_idx,
                                    net_name=net_name,
                                    share_name=share_name,
                                    ip_address=ip_address,
                                    client_computer=client_computer,
                                    new_node=node_idx)
                _check_generic_move(regs,
                                    apply_to_all=apply_to_all,
                                    registration_idx=registration_idx,
                                    net_name=net_name,
                                    share_name=share_name,
                                    ip_address=ip_address,
                                    client_computer=client_computer,
                                    new_ip=node['ip'])

        if msg_type == witness.WITNESS_NOTIFY_CLIENT_MOVE:
            only_shares = False
        elif msg_type == witness.WITNESS_NOTIFY_SHARE_MOVE:
            only_shares = True

        self.check_combinations(check_generic_move, only_shares=only_shares)

    def test_net_witness_client_move(self):
        self._test_net_witness_generic_move('client-move',
                                            'CLIENT_MOVE_TO',
                                            witness.WITNESS_NOTIFY_CLIENT_MOVE)
    def test_net_witness_share_move(self):
        self._test_net_witness_generic_move('share-move',
                                            'SHARE_MOVE_TO',
                                            witness.WITNESS_NOTIFY_SHARE_MOVE)

    def test_net_witness_force_unregister(self):
        def check_force_unregister(regs,
                                   apply_to_all=False,
                                   registration_idx=None,
                                   net_name=None,
                                   share_name=None,
                                   ip_address=None,
                                   client_computer=None):
            def check_force_unregister_happened(reg):
                conn = reg['conn']
                reg_context = reg['context']
                self.assertIsNotNone(reg_context)
                try:
                    conn.UnRegister(reg_context)
                    self.fail()
                except WERRORError as e:
                    (num, string) = e.args
                    if num != werror.WERR_NOT_FOUND:
                        raise
                reg['context'] = None

            return self.check_net_witness_output("force-unregister",
                                                 regs,
                                                 apply_to_all=apply_to_all,
                                                 registration_idx=registration_idx,
                                                 net_name=net_name,
                                                 share_name=share_name,
                                                 ip_address=ip_address,
                                                 client_computer=client_computer,
                                                 expected_msg_type="FORCE_UNREGISTER",
                                                 callback=check_force_unregister_happened)

        self.check_combinations(check_force_unregister)

    def _test_net_witness_force_response(self,
                                         msg_type=None,
                                         expected_resource_changes=None,
                                         expected_ip_lists=None):
        def check_force_response(regs,
                                 apply_to_all=False,
                                 registration_idx=None,
                                 net_name=None,
                                 share_name=None,
                                 ip_address=None,
                                 client_computer=None):
            move_types = [
                witness.WITNESS_NOTIFY_CLIENT_MOVE,
                witness.WITNESS_NOTIFY_SHARE_MOVE,
                witness.WITNESS_NOTIFY_IP_CHANGE,
            ]

            forced_response = '{ '
            forced_response +=   '"result": 0, '
            forced_response +=   '"response": { '
            forced_response +=     '"type": %u, ' % msg_type
            forced_response +=     '"messages": [ '
            if msg_type == witness.WITNESS_NOTIFY_RESOURCE_CHANGE:
                prefix_d1 = ""
                for rc in expected_resource_changes:
                    forced_response += prefix_d1
                    forced_response += '{ '
                    prefix_d2 = ""
                    if 'type' in rc:
                        forced_response += prefix_d2
                        forced_response += '"type": %u ' % rc['type']
                        prefix_d2 = ", "
                    if 'name' in rc:
                        forced_response += prefix_d2
                        forced_response += '"name": "%s" ' % rc['name']
                        prefix_d2 = ", "
                    forced_response += '} '
                    prefix_d1 = ", "
            if msg_type in move_types:
                prefix_d1 = ""
                for ip_list in expected_ip_lists:
                    forced_response += prefix_d1
                    forced_response += '['
                    prefix_d2 = ""
                    for ip in ip_list:
                        forced_response += prefix_d2
                        forced_response +=   '{ '
                        prefix_d3 = ""
                        if 'flags' in ip:
                            forced_response += prefix_d3
                            forced_response += '"flags": %u' % ip['flags']
                            prefix_d3 = ", "
                        if 'ipv4' in ip:
                            forced_response += prefix_d3
                            forced_response += '"ipv4": "%s" ' % ip['ipv4']
                            prefix_d3 = ", "
                        if 'ipv6' in ip:
                            forced_response += prefix_d3
                            forced_response += '"ipv6": "%s" ' % ip['ipv6']
                            prefix_d3 = ", "
                        forced_response +=   '}'
                        prefix_d2 = ", "
                    forced_response += ']'
                    prefix_d1 = ", "
            forced_response +=     ']'
            forced_response +=   '}'
            forced_response += '}'

            def check_forced_response_result(reg):
                conn = reg['conn']
                reg_context = reg['context']
                response = conn.AsyncNotify(reg_context)
                if msg_type == witness.WITNESS_NOTIFY_RESOURCE_CHANGE:
                    self.assertResourceChanges(response, expected_resource_changes)
                if msg_type in move_types:
                    self.assertGenericIpLists(response, msg_type, expected_ip_lists)

            return self.check_net_witness_output("force-response",
                                                 regs,
                                                 apply_to_all=apply_to_all,
                                                 registration_idx=registration_idx,
                                                 net_name=net_name,
                                                 share_name=share_name,
                                                 ip_address=ip_address,
                                                 client_computer=client_computer,
                                                 forced_response=forced_response,
                                                 expected_msg_type="FORCE_RESPONSE",
                                                 callback=check_forced_response_result)

        self.check_combinations(check_force_response)

    def test_net_witness_force_response_resource_changes(self):
        msg_type = witness.WITNESS_NOTIFY_RESOURCE_CHANGE
        expected_resource_changes = [
            {
                'type': witness.WITNESS_RESOURCE_STATE_UNAVAILABLE,
                'name': "some-resource-name"
            },
            {
                'type': witness.WITNESS_RESOURCE_STATE_AVAILABLE,
                'name': "other-resource-name"
            },
        ]
        self._test_net_witness_force_response(msg_type=msg_type,
                                              expected_resource_changes=expected_resource_changes)

    def _test_net_witness_force_response_generic_moves(self, msg_type):
        expected_flags = 0
        expected_flags |= witness.WITNESS_IPADDR_V4
        expected_flags |= witness.WITNESS_IPADDR_ONLINE

        expected_ip_list10 = [
            {
                'flags': expected_flags,
                'ipv4': '10.0.10.1',
            },
            {
                'flags': 0,
                'ipv4': '10.0.10.2',
                'ipv6': 'fd00:0000:0000:0000:0010:0000:0010:0002',
            },
        ]
        expected_ip_list20 = [
            {
                'flags': expected_flags,
                'ipv4': '10.0.20.1',
            },
            {
                'flags': 0,
                'ipv4': '10.0.20.2',
                'ipv6': 'fd00:0000:0000:0000:0010:0000:0020:0002',
            },
        ]

        expected_ip_lists = [expected_ip_list10, expected_ip_list20]
        self._test_net_witness_force_response(msg_type=msg_type,
                                              expected_ip_lists=expected_ip_lists)

    def test_net_witness_force_response_client_moves(self):
        msg_type = witness.WITNESS_NOTIFY_CLIENT_MOVE
        self._test_net_witness_force_response_generic_moves(msg_type)

    def test_net_witness_force_response_share_moves(self):
        msg_type = witness.WITNESS_NOTIFY_SHARE_MOVE
        self._test_net_witness_force_response_generic_moves(msg_type)

    def test_net_witness_force_response_ip_changes(self):
        msg_type = witness.WITNESS_NOTIFY_IP_CHANGE
        self._test_net_witness_force_response_generic_moves(msg_type)

if __name__ == "__main__":
    import unittest
    unittest.main()



# Unix SMB/CIFS implementation.
#
# authentication policy - manage computer-allowed-to-authenticate-to property
#
# Copyright (C) Catalyst.Net Ltd. 2024
#
# Written by Rob van der Linde <rob@catalyst.net.nz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from samba.domain.models import AuthenticationPolicy, AuthenticationSilo, Group
from samba.domain.models.exceptions import ModelError
from samba.getopt import CredentialsOptions, HostOptions, Option, SambaOptions
from samba.netcmd import Command, CommandError, SuperCommand


class cmd_domain_auth_policy_computer_allowed_to_authenticate_to_set(Command):
    """Set the computer-allowed-to-authenticate-to property based on scenario.

    --by-group:

        The computer account (server, workstation) service requires the
        connecting user to be in GROUP.

    --by-silo:

        The computer account (server, workstation) service requires the
        connecting user to be in SILO.

    The options above are mutually exclusive, only one can be set at a time.
    """

    synopsis = "%prog -H <URL> [options]"

    takes_optiongroups = {
        "sambaopts": SambaOptions,
        "credopts": CredentialsOptions,
        "hostopts": HostOptions,
    }

    takes_options = [
        Option("--name",
               help="Name of authentication policy to view (required).",
               dest="name", action="store", type=str, required=True),
        Option("--by-group",
               help="The computer account (server, workstation) service "
                    "requires the connecting user to be in GROUP.",
               dest="groupname", action="store", type=str),
        Option("--by-silo",
               help="The computer account (server, workstation) service "
                    "requires the connecting user to be in SILO.",
               dest="siloname", action="store", type=str),
    ]

    def run(self, hostopts=None, sambaopts=None, credopts=None, name=None,
            groupname=None, siloname=None):

        if groupname and siloname:
            raise CommandError("Cannot have both --by-group and --by-silo options.")

        ldb = self.ldb_connect(hostopts, sambaopts, credopts)

        try:
            policy = AuthenticationPolicy.get(ldb, cn=name)
        except ModelError as e:
            raise CommandError(e)

        if policy is None:
            raise CommandError(f"Authentication policy {name} not found.")

        if groupname:
            try:
                group = Group.get(ldb, cn=groupname)
            except ModelError as e:
                raise CommandError(e)

            if group is None:
                raise CommandError(f"Group {groupname} not found.")

            sddl = group.get_authentication_sddl()

        elif siloname:
            try:
                silo = AuthenticationSilo.get(ldb, cn=siloname)
            except ModelError as e:
                raise CommandError(e)

            if silo is None:
                raise CommandError(f"Authentication silo {siloname} not found.")

            sddl = silo.get_authentication_sddl()

        else:
            raise CommandError("Either --by-group or --by-silo expected.")

        policy.computer_allowed_to_authenticate_to = sddl

        try:
            policy.save(ldb)
        except ModelError as e:
            raise CommandError(e)

        # Authentication policy updated successfully.
        print(f"Updated authentication policy: {name}", file=self.outf)
        print(f"Updated SDDL: {sddl}", file=self.outf)


class cmd_domain_auth_policy_computer_allowed_to_authenticate_to(SuperCommand):
    """Manage the computer-allowed-to-authenticate-to property."""

    subcommands = {
        "set": cmd_domain_auth_policy_computer_allowed_to_authenticate_to_set(),
    }



# Unix SMB/CIFS implementation.
#
# Copyright (C) 2017      Andreas Schneider <asn@samba.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import samba.tests
import pypamtest
import os


class SimplePamTests(samba.tests.TestCase):
    def test_authenticate(self):
        domain = os.environ["DOMAIN"]
        username = os.environ["USERNAME"]
        password = os.environ["PASSWORD"]
        if domain != "":
            unix_username = "%s/%s" % (domain, username)
        else:
            unix_username = "%s" % username
        expected_rc = 0  # PAM_SUCCESS

        tc = pypamtest.TestCase(pypamtest.PAMTEST_AUTHENTICATE, expected_rc)
        try:
            res = pypamtest.run_pamtest(unix_username, "samba", [tc], [password])
        except pypamtest.PamTestError as e:
            raise AssertionError(str(e))

        self.assertTrue(res is not None)

    def test_authenticate_error(self):
        domain = os.environ["DOMAIN"]
        username = os.environ["USERNAME"]
        password = "WrongPassword"
        if domain != "":
            unix_username = "%s/%s" % (domain, username)
        else:
            unix_username = "%s" % username
        expected_rc = 7  # PAM_AUTH_ERR

        tc = pypamtest.TestCase(pypamtest.PAMTEST_AUTHENTICATE, expected_rc)
        try:
            res = pypamtest.run_pamtest(unix_username, "samba", [tc], [password])
        except pypamtest.PamTestError as e:
            raise AssertionError(str(e))

        self.assertTrue(res is not None)

        # Authenticate again to check that we are not locked out with just one
        # failed login
        password = os.environ["PASSWORD"]
        expected_rc = 0  # PAM_SUCCESS

        tc = pypamtest.TestCase(pypamtest.PAMTEST_AUTHENTICATE, expected_rc)
        try:
            res = pypamtest.run_pamtest(unix_username, "samba", [tc], [password])
        except pypamtest.PamTestError as e:
            raise AssertionError(str(e))

        self.assertTrue(res is not None)



# Unix SMB/CIFS implementation.
# Copyright (C) Catalyst IT Ltd. 2017
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from samba.tests.samba_tool.base import SambaToolCmdTest
import os
import shutil


class ProvisionPasswordTestCase(SambaToolCmdTest):
    """Test for password validation in domain provision subcommand"""

    def setUp(self):
        super().setUp()
        self.tempsambadir = os.path.join(self.tempdir, "samba")
        os.mkdir(self.tempsambadir)

    def _provision_with_password(self, password):
        return self.runsubcmd(
            "domain", "provision", "--realm=foo.example.com", "--domain=FOO",
            "--targetdir=%s" % self.tempsambadir, "--adminpass=%s" % password,
            "--use-ntvfs")

    def test_short_and_low_quality(self):
        (result, out, err) = self._provision_with_password("foo")
        self.assertCmdFail(result)

    def test_short(self):
        (result, out, err) = self._provision_with_password("Fo0!_9")
        self.assertCmdFail(result)
        self.assertRegex(err, r"minimum password length")

    def test_low_quality(self):
        (result, out, err) = self._provision_with_password("aaaaaaaaaaaaaaaaa")
        self.assertCmdFail(result)
        self.assertRegex(err, r"quality standards")

    def test_good(self):
        (result, out, err) = self._provision_with_password("Fo0!_9.")
        self.assertCmdSuccess(result, out, err)

    def tearDown(self):
        super().tearDown()
        shutil.rmtree(self.tempsambadir)



# user management
#
# add user
#
# Copyright Jelmer Vernooij 2010 <jelmer@samba.org>
# Copyright Theresa Halloran 2011 <theresahalloran@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import pwd
from getpass import getpass

import samba.getopt as options
from samba import generate_random_password
from samba.auth import system_session
from samba.netcmd import Command, CommandError, Option
from samba.samdb import SamDB


class cmd_user_add(Command):
    """Add a new user.

This command adds a new user account to the Active Directory domain.  The username specified on the command is the sAMaccountName.

User accounts may represent physical entities, such as people or may be used as service accounts for applications.  User accounts are also referred to as security principals and are assigned a security identifier (SID).

A user account enables a user to logon to a computer and domain with an identity that can be authenticated.  To maximize security, each user should have their own unique user account and password.  A user's access to domain resources is based on permissions assigned to the user account.

Unix (RFC2307) attributes may be added to the user account. Attributes taken from NSS are obtained on the local machine. Explicitly given values override values obtained from NSS. Configure 'idmap_ldb:use rfc2307 = Yes' to use these attributes for UID/GID mapping.

The command may be run from the root userid or another authorized userid.  The -H or --URL= option can be used to execute the command against a remote server.

Example1:
samba-tool user add User1 passw0rd --given-name=John --surname=Smith --must-change-at-next-login -H ldap://samba.samdom.example.com -Uadministrator%passw1rd

Example1 shows how to add a new user to the domain against a remote LDAP server.  The -H parameter is used to specify the remote target server.  The -U option is used to pass the userid and password authorized to issue the command remotely.

Example2:
sudo samba-tool user add User2 passw2rd --given-name=Jane --surname=Doe --must-change-at-next-login

Example2 shows how to add a new user to the domain against the local server.   sudo is used so a user may run the command as root.  In this example, after User2 is created, he/she will be forced to change their password when they logon.

Example3:
samba-tool user add User3 passw3rd --userou='OU=OrgUnit'

Example3 shows how to add a new user in the OrgUnit organizational unit.

Example4:
samba-tool user add User4 passw4rd --rfc2307-from-nss --gecos 'some text'

Example4 shows how to add a new user with Unix UID, GID and login-shell set from the local NSS and GECOS set to 'some text'.

Example5:
samba-tool user add User5 passw5rd --nis-domain=samdom --unix-home=/home/User5 \\
    --uid-number=10005 --login-shell=/bin/false --gid-number=10000

Example5 shows how to add a new RFC2307/NIS domain enabled user account. If
--nis-domain is set, then the other four parameters are mandatory.

"""
    synopsis = "%prog <username> [<password>] [options]"

    takes_options = [
        Option("-H", "--URL", help="LDB URL for database or target server", type=str,
               metavar="URL", dest="H"),
        Option("--must-change-at-next-login",
               help="Force password to be changed on next login",
               action="store_true"),
        Option("--random-password",
               help="Generate random password",
               action="store_true"),
        Option("--smartcard-required",
               help="Require a smartcard for interactive logons",
               action="store_true"),
        Option("--use-username-as-cn",
               help="Force use of username as user's CN",
               action="store_true"),
        Option("--userou",
               help="DN of alternative location (without domainDN counterpart) to default CN=Users in which new user object will be created. E. g. 'OU=<OU name>'",
               type=str),
        Option("--surname", help="User's surname", type=str),
        Option("--given-name", help="User's given name", type=str),
        Option("--initials", help="User's initials", type=str),
        Option("--profile-path", help="User's profile path", type=str),
        Option("--script-path", help="User's logon script path", type=str),
        Option("--home-drive", help="User's home drive letter", type=str),
        Option("--home-directory", help="User's home directory path", type=str),
        Option("--job-title", help="User's job title", type=str),
        Option("--department", help="User's department", type=str),
        Option("--company", help="User's company", type=str),
        Option("--description", help="User's description", type=str),
        Option("--mail-address", help="User's email address", type=str),
        Option("--internet-address", help="User's home page", type=str),
        Option("--telephone-number", help="User's phone number", type=str),
        Option("--physical-delivery-office", help="User's office location", type=str),
        Option("--rfc2307-from-nss",
               help="Copy Unix user attributes from NSS (will be overridden by explicit UID/GID/GECOS/shell)",
               action="store_true"),
        Option("--nis-domain", help="User's Unix/RFC2307 NIS domain", type=str),
        Option("--unix-home", help="User's Unix/RFC2307 home directory",
               type=str),
        Option("--uid", help="User's Unix/RFC2307 username", type=str),
        Option("--uid-number", help="User's Unix/RFC2307 numeric UID", type=int),
        Option("--gid-number", help="User's Unix/RFC2307 primary GID number", type=int),
        Option("--gecos", help="User's Unix/RFC2307 GECOS field", type=str),
        Option("--login-shell", help="User's Unix/RFC2307 login shell", type=str),
    ]

    takes_args = ["username", "password?"]

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "credopts": options.CredentialsOptions,
        "versionopts": options.VersionOptions,
    }

    def run(self, username, password=None, credopts=None, sambaopts=None,
            versionopts=None, H=None, must_change_at_next_login=False,
            random_password=False, use_username_as_cn=False, userou=None,
            surname=None, given_name=None, initials=None, profile_path=None,
            script_path=None, home_drive=None, home_directory=None,
            job_title=None, department=None, company=None, description=None,
            mail_address=None, internet_address=None, telephone_number=None,
            physical_delivery_office=None, rfc2307_from_nss=False,
            nis_domain=None, unix_home=None, uid=None, uid_number=None,
            gid_number=None, gecos=None, login_shell=None,
            smartcard_required=False):

        if smartcard_required:
            if password is not None and password != '':
                raise CommandError('It is not allowed to specify '
                                   '--newpassword '
                                   'together with --smartcard-required.')
            if must_change_at_next_login:
                raise CommandError('It is not allowed to specify '
                                   '--must-change-at-next-login '
                                   'together with --smartcard-required.')

        if random_password and not smartcard_required:
            password = generate_random_password(128, 255)

        while True:
            if smartcard_required:
                break
            if password is not None and password != '':
                break
            password = getpass("New Password: ")
            passwordverify = getpass("Retype Password: ")
            if not password == passwordverify:
                password = None
                self.outf.write("Sorry, passwords do not match.\n")

        if rfc2307_from_nss:
            pwent = pwd.getpwnam(username)
            if uid is None:
                uid = username
            if uid_number is None:
                uid_number = pwent[2]
            if gid_number is None:
                gid_number = pwent[3]
            if gecos is None:
                gecos = pwent[4]
            if login_shell is None:
                login_shell = pwent[6]

        lp = sambaopts.get_loadparm()
        creds = credopts.get_credentials(lp)

        if uid_number or gid_number:
            if not lp.get("idmap_ldb:use rfc2307"):
                self.outf.write("You are setting a Unix/RFC2307 UID or GID. You may want to set 'idmap_ldb:use rfc2307 = Yes' to use those attributes for XID/SID-mapping.\n")

        if nis_domain is not None:
            if None in (uid_number, login_shell, unix_home, gid_number):
                raise CommandError('Missing parameters. To enable NIS features, '
                                   'the following options have to be given: '
                                   '--nis-domain=, --uid-number=, --login-shell='
                                   ', --unix-home=, --gid-number= Operation '
                                   'cancelled.')

        try:
            samdb = SamDB(url=H, session_info=system_session(),
                          credentials=creds, lp=lp)
            samdb.newuser(username, password, force_password_change_at_next_login_req=must_change_at_next_login,
                          useusernameascn=use_username_as_cn, userou=userou, surname=surname, givenname=given_name, initials=initials,
                          profilepath=profile_path, homedrive=home_drive, scriptpath=script_path, homedirectory=home_directory,
                          jobtitle=job_title, department=department, company=company, description=description,
                          mailaddress=mail_address, internetaddress=internet_address,
                          telephonenumber=telephone_number, physicaldeliveryoffice=physical_delivery_office,
                          nisdomain=nis_domain, unixhome=unix_home, uid=uid,
                          uidnumber=uid_number, gidnumber=gid_number,
                          gecos=gecos, loginshell=login_shell,
                          smartcard_required=smartcard_required)
        except Exception as e:
            raise CommandError("Failed to add user '%s': " % username, e)

        self.outf.write("User '%s' added successfully\n" % username)



# Unix SMB/CIFS implementation.
# Copyright (C) Volker Lendecke <vl@samba.org> 2021
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""These tests compare Windows security descriptors with Samba
descriptors derived from the same SDDL.

They use json and json.gz files in libcli/security/tests/data.
"""

from samba.dcerpc import security
from samba.ndr import ndr_pack, ndr_unpack, ndr_print
from samba.tests import TestCase, DynamicTestCase
from samba.colour import colourdiff
from hashlib import md5
import gzip

import json
from pathlib import Path

TEST_DIR = Path(__name__).parent.parent.parent / 'libcli/security/tests/data'


class SDDLvsDescriptorBase(TestCase):
    """These tests have no explicit cases and no inline data. The actual
    data is kept in JSON files in libcli/security/tests/data, so that
    it easy to share those files with Windows. To control what tests
    are run, set the `json_file` attribute in subclasses, and/or add a
    filter_test_cases class method.
    """
    maxDiff = 10000
    json_file = None
    munge_to_v4 = True
    domain_sid = security.dom_sid("S-1-5-21-2457507606-2709100691-398136650")
    failure_json = None
    success_json = None

    @classmethod
    def filter_test_cases(cls, data):
        """Filter out some cases before running the tests.
        Like this, for example:
            return {k:v for k, v in data.items() if len(k) < 200 and
                    '(D;;;;;MP)(D;;;;;MP)(D;;;;;MP)' in k}
        """
        return data

    @classmethod
    def setUpDynamicTestCases(cls):
        try:
            with gzip.open(cls.json_file, 'rt') as f:
                data = json.load(f)
        except Exception:
            with open(cls.json_file) as f:
                data = json.load(f)

        data = cls.filter_test_cases(data)
        i = 0
        for sddl, sdl in data.items():
            i += 1
            name = f'{i:03}-{sddl}'
            if len(name) > 130:
                tag = md5(sddl.encode()).hexdigest()[:10]
                name = f"{name[:100]}+{len(name) - 100}-more-characters-{tag}"
            cls.generate_dynamic_test('test_sddl_vs_sd', name, sddl, sdl)

        if cls.failure_json:
            cls.failures = {}
            cls.failure_file = open(cls.failure_json, 'w')
            cls.addClassCleanup(json.dump, cls.failures, cls.failure_file)
        if cls.success_json:
            cls.successes = {}
            cls.success_file = open(cls.success_json, 'w')
            cls.addClassCleanup(json.dump, cls.successes, cls.success_file)

    def _test_sddl_vs_sd_with_args(self, sddl, sdl):
        sdb_win = bytes(sdl)
        try:
            sd_sam = security.descriptor.from_sddl(sddl, self.domain_sid)
        except (TypeError, ValueError, security.SDDLValueError) as e:
            try:
                sd_win = ndr_unpack(security.descriptor, sdb_win)
                win_ndr_print = ndr_print(sd_win)
            except RuntimeError as e2:
                win_ndr_print = f"not parseable: {e2}"
            if self.failure_json:
                self.failures[sddl] = sdl

            self.fail(f"failed to parse {sddl} into SD: {e}")

        try:
            sdb_sam = ndr_pack(sd_sam)
        except RuntimeError as e:
            if self.failure_json:
                self.failures[sddl] = sdl
            self.fail(f"failed to pack samba SD from {sddl} into bytes: {e}\n"
                      f"{ndr_print(sd_sam)}")

        try:
            sd_win = ndr_unpack(security.descriptor, sdb_win)
        except RuntimeError as e:
            if self.failure_json:
                self.failures[sddl] = sdl
            self.fail(f"could not unpack windows descriptor for {sddl}: {e}")

        if self.munge_to_v4:
            # Force the ACL revisions to match Samba. Windows seems to
            # use the lowest possible revision, while Samba uses
            # ACL_REVISION_DS when generating from SDDL. The _DS
            # version allows more ACE types, but is otherwise the same.
            #
            # MS-DTYP 2.4.5 ACL:
            #
            # ACL_REVISION 0x02
            #
            # When set to 0x02, only AceTypes 0x00, 0x01,
            # 0x02, 0x03, 0x11, 0x12, and 0x13 can be present in the ACL.
            # An AceType of 0x11 is used for SACLs but not for DACLs. For
            # more information about ACE types, see section 2.4.4.1.
            #
            # ACL_REVISION_DS 0x04
            #
            # When set to 0x04, AceTypes 0x05, 0x06, 0x07, 0x08, and 0x11
            # are allowed. ACLs of revision 0x04 are applicable only to
            # directory service objects. An AceType of 0x11 is used for
            # SACLs but not for DACLs.
            #
            # 5, 6, 7, 8 are object ACES.
            if sd_win.dacl:
                sd_win.dacl.revision = 4
            if sd_win.sacl:
                sd_win.sacl.revision = 4

        if (sd_win != sd_sam):
            if self.failure_json:
                self.failures[sddl] = sdl
            self.fail(f"Descriptors differ for {sddl}")

        if self.success_json:
            self.successes[sddl] = sdl


@DynamicTestCase
class SDDLvsDescriptorShortOrdinaryAcls(SDDLvsDescriptorBase):
    """These are not conditional ACEs or resource attribute aces, the SDDL
    is less than 1000 characters long, and success is expected.
    """
    json_file = TEST_DIR / 'short-ordinary-acls.json.gz'


@DynamicTestCase
class SDDLvsDescriptorRegistryObjectRights(SDDLvsDescriptorBase):
    """We'll fail these because we don't recognise 'KA' and related object
    rights strings that are used for registry objects."""
    json_file = TEST_DIR / 'registry-object-rights.json'


@DynamicTestCase
class SDDLvsDescriptorOverSizeAcls(SDDLvsDescriptorBase):
    """These are ordinary ACLs that contain duplicate ACEs (e.g.
    'D:P(D;;;;;MP)(D;;;;;MP)(D;;;;;MP)(D;;;;;MP)'). Due to a
    peculiarity in Windows, the ACL structures generated have extra
    trailing zero bytes. Due to a peculiarity in the way Samba reads
    an ACL (namely, it assumes an ACL will be just big enough for its
    ACEs), these cannot currently be parsed by Samba.
    """
    json_file = TEST_DIR / 'oversize-acls.json'


@DynamicTestCase
class SDDLvsDescriptorShortConditionalAndResourceAceSuccesses(SDDLvsDescriptorBase):
    """These contain conditional ACEs or resource attribute aces, the SDDL
    is less than 1000 characters long, and success is expected.
    """
    json_file = TEST_DIR / 'short-conditional-and-resource-aces-successes.json.gz'


@DynamicTestCase
class SDDLvsDescriptorShortConditionalAndResourceAcesTxIntegers(SDDLvsDescriptorBase):
    """These contain resource attribute aces in the form

          (RA;;;;;WD;("foo",TX,0x0,0077,00,...))

    where the numbers after the 0x0 flags like "0077" are interpreted
    by Windows as if they are octet strings. This is not documented
    and not supported by Samba.
    """
    json_file = TEST_DIR / 'short-conditional-and-resource-aces-tx-int.json.gz'


@DynamicTestCase
class SDDLvsDescriptorShortOrdinaryAclsNoMungeV4(SDDLvsDescriptorBase):
    """These ones have revision 2 ACLs (NT4), but Samba's SDDL only writes
    revision 4 ACLs (which are otherwise identical).
    """
    munge_to_v4 = False
    json_file = TEST_DIR / 'short-ordinary-acls-v2.json.gz'


@DynamicTestCase
class SDDLvsDescriptorCollectedConditionalAces(SDDLvsDescriptorBase):
    """Some conditional ACE strings that have collected up.
    """
    json_file = TEST_DIR / 'conditional_aces.txt.json'
